from . import *
from dataclasses import dataclass, asdict, field
from typing import Dict, Union, Optional, Any, ClassVar, TypeVar


# common keys used in the format configuration
_FORMAT_NAME_KEY = "formatName"


@dataclass
class _FormatBase(object):
    # abstract variables that need to be implemented by Derived Classes
    name: ClassVar[str]

    # methods to be implemented by every Format
    def _build(self) -> dict:
        """
        abstract method that needs to be implemented by derived class.

        :return: Configuration properties in a dictionary that will be used to make the Rest call to the backend application
        """
        raise NotImplementedError

    @classmethod
    def _from_config(cls, config: dict) -> object:
        raise NotImplementedError


@dataclass
class GeoRssFormat(_FormatBase):
    T = TypeVar("GeoRssFormat")
    name: ClassVar[str] = "geo-rss-format"

    # format keys
    _DATE_FORMAT_KEY: str = field(init=False, default=f"{name}.dateFormat")

    # this format's instance properties
    date_format: Optional[str] = None

    def _build(self) -> dict:
        result_dict = {_FORMAT_NAME_KEY: GeoRssFormat.name}
        if self.date_format is not None:
            result_dict |= {
                "properties": {
                    f"{GeoRssFormat.name}.{GeoRssFormat._DATE_FORMAT_KEY}": self.date_format
                }
            }

        return result_dict

    @classmethod
    def _from_config(cls, config: dict) -> T:
        properties = config["properties"]
        obj = cls(date_format=properties.get(GeoRssFormat._DATE_FORMAT_KEY))
        return obj


@dataclass
class RssFormat(_FormatBase):
    T = TypeVar("RssFormat")
    name: ClassVar[str] = "rss-format"

    # format keys
    _BUILD_GEOMETRY_FROM_FIELDS_KEY: str = field(
        init=False, default=f"{name}.buildGeometryFromFields"
    )
    _X_FIELD_KEY: str = field(init=False, default=f"{name}.xField")
    _Y_FIELD_KEY: str = field(init=False, default=f"{name}.yField")
    _HAS_Z_FIELD_KEY_KEY: str = field(init=False, default=f"{name}.hasZField")
    _Z_FIELD_KEY: str = field(init=False, default=f"{name}.zField")
    _Z_UNIT_KEY: str = field(init=False, default=f"{name}.zUnit")
    _GEOMETRY_FIELD_KEY: str = field(init=False, default=f"{name}.geometryField")
    _GEOMETRY_FIELD_FORMAT_KEY: str = field(
        init=False, default=f"{name}.geometryFieldFormat"
    )

    _DATE_FORMAT_KEY: str = field(init=False, default=f"{name}.dateFormat")

    # this format's instance properties
    # date properties
    date_format: Optional[str] = None
    # geometry properties
    build_geometry_from_fields: Optional[bool] = None
    x_field: Optional[str] = None
    y_field: Optional[str] = None
    has_z_field: Optional[bool] = None
    z_field: Optional[str] = None
    z_unit: Optional[int] = None
    geometry_field: Optional[str] = None
    geometry_field_format: Optional[str] = None

    def _build(self) -> dict:
        result_dict = {_FORMAT_NAME_KEY: self.name}
        properties_dict = {}

        if self.build_geometry_from_fields is not None:
            properties_dict[
                RssFormat._BUILD_GEOMETRY_FROM_FIELDS_KEY
            ] = self.build_geometry_from_fields
        if self.x_field is not None:
            properties_dict[RssFormat._X_FIELD_KEY] = self.x_field
        if self.y_field is not None:
            properties_dict[RssFormat._Y_FIELD_KEY] = self.y_field
        if self.has_z_field is not None:
            properties_dict[RssFormat._HAS_Z_FIELD_KEY_KEY] = self.has_z_field
        if self.z_field is not None:
            properties_dict[RssFormat._Z_FIELD_KEY] = self.z_field
        if self.z_unit is not None:
            properties_dict[RssFormat._Z_UNIT_KEY] = self.z_unit
        if self.geometry_field is not None:
            properties_dict[RssFormat._GEOMETRY_FIELD_KEY] = self.geometry_field
        if self.geometry_field_format is not None:
            properties_dict[
                RssFormat._GEOMETRY_FIELD_FORMAT_KEY
            ] = self.geometry_field_format
        if self.date_format is not None:
            properties_dict[RssFormat._DATE_FORMAT_KEY] = self.date_format

        if bool(properties_dict):
            result_dict["properties"] = properties_dict

        return result_dict

    @classmethod
    def _from_config(cls, config: dict) -> T:
        properties = config["properties"]
        obj = RssFormat(
            build_geometry_from_fields=properties.get(
                RssFormat._BUILD_GEOMETRY_FROM_FIELDS_KEY
            ),
            x_field=properties.get(RssFormat._X_FIELD_KEY),
            y_field=properties.get(RssFormat._Y_FIELD_KEY),
            has_z_field=properties.get(RssFormat._HAS_Z_FIELD_KEY_KEY),
            z_field=properties.get(RssFormat._Z_FIELD_KEY),
            z_unit=properties.get(RssFormat._Z_UNIT_KEY),
            geometry_field=properties.get(RssFormat._GEOMETRY_FIELD_KEY),
            geometry_field_format=properties.get(RssFormat._GEOMETRY_FIELD_FORMAT_KEY),
            date_format=properties.get(RssFormat._DATE_FORMAT_KEY),
        )

        return obj


@dataclass
class DelimitedFormat(_FormatBase):
    T = TypeVar("DelimitedFormat")
    name: ClassVar[str] = "delimited"

    # format keys
    _BUILD_GEOMETRY_FROM_FIELDS_KEY: str = field(
        init=False, default=f"{name}.buildGeometryFromFields"
    )
    _X_FIELD_KEY: str = field(init=False, default=f"{name}.xField")
    _Y_FIELD_KEY: str = field(init=False, default=f"{name}.yField")
    _HAS_Z_FIELD_KEY_KEY: str = field(init=False, default=f"{name}.hasZField")
    _Z_FIELD_KEY: str = field(init=False, default=f"{name}.zField")
    _Z_UNIT_KEY: str = field(init=False, default=f"{name}.zUnit")
    _GEOMETRY_FIELD_KEY: str = field(init=False, default=f"{name}.geometryField")
    _GEOMETRY_FIELD_FORMAT_KEY: str = field(
        init=False, default=f"{name}.geometryFieldFormat"
    )

    _DATE_FORMAT_KEY: str = field(init=False, default=f"{name}.dateFormat")

    _RECORD_TERMINATOR_KEY: str = field(init=False, default=f"{name}.recordTerminator")
    _FIELD_DELIMITER_KEY: str = field(init=False, default=f"{name}.fieldDelimiter")
    _HAS_HEADER_ROW: str = field(init=False, default=f"{name}.hasHeaderRow")

    # this format's instance properties
    record_terminator: Optional[str] = None
    field_delimiter: Optional[str] = None
    has_header_row: Optional[bool] = None
    # date properties
    date_format: Optional[str] = None
    # geometry properties
    build_geometry_from_fields: Optional[bool] = None
    x_field: Optional[str] = None
    y_field: Optional[str] = None
    has_z_field: Optional[bool] = None
    z_field: Optional[str] = None
    z_unit: Optional[int] = None
    geometry_field: Optional[str] = None
    geometry_field_format: Optional[str] = None

    def _build(self) -> dict:
        result_dict = {_FORMAT_NAME_KEY: self.name}
        properties_dict = {}

        if self.build_geometry_from_fields is not None:
            properties_dict[
                DelimitedFormat._BUILD_GEOMETRY_FROM_FIELDS_KEY
            ] = self.build_geometry_from_fields
        if self.x_field is not None:
            properties_dict[DelimitedFormat._X_FIELD_KEY] = self.x_field
        if self.y_field is not None:
            properties_dict[DelimitedFormat._Y_FIELD_KEY] = self.y_field
        if self.has_z_field is not None:
            properties_dict[DelimitedFormat._HAS_Z_FIELD_KEY_KEY] = self.has_z_field
        if self.z_field is not None:
            properties_dict[DelimitedFormat._Z_FIELD_KEY] = self.z_field
        if self.z_unit is not None:
            properties_dict[DelimitedFormat._Z_UNIT_KEY] = self.z_unit
        if self.geometry_field is not None:
            properties_dict[DelimitedFormat._GEOMETRY_FIELD_KEY] = self.geometry_field
        if self.geometry_field_format is not None:
            properties_dict[
                DelimitedFormat._GEOMETRY_FIELD_FORMAT_KEY
            ] = self.geometry_field_format
        if self.date_format is not None:
            properties_dict[DelimitedFormat._DATE_FORMAT_KEY] = self.date_format

        if self.record_terminator is not None:
            properties_dict[
                DelimitedFormat._RECORD_TERMINATOR_KEY
            ] = self.record_terminator
        if self.field_delimiter is not None:
            properties_dict[DelimitedFormat._FIELD_DELIMITER_KEY] = self.field_delimiter
        if self.has_header_row is not None:
            properties_dict[DelimitedFormat._HAS_HEADER_ROW] = self.has_header_row

        if bool(properties_dict):
            result_dict["properties"] = properties_dict

        return result_dict

    @classmethod
    def _from_config(cls, config: dict) -> T:
        properties = config["properties"]
        obj = DelimitedFormat(
            build_geometry_from_fields=properties.get(
                DelimitedFormat._BUILD_GEOMETRY_FROM_FIELDS_KEY
            ),
            x_field=properties.get(DelimitedFormat._X_FIELD_KEY),
            y_field=properties.get(DelimitedFormat._Y_FIELD_KEY),
            has_z_field=properties.get(DelimitedFormat._HAS_Z_FIELD_KEY_KEY),
            z_field=properties.get(DelimitedFormat._Z_FIELD_KEY),
            z_unit=properties.get(DelimitedFormat._Z_UNIT_KEY),
            geometry_field=properties.get(DelimitedFormat._GEOMETRY_FIELD_KEY),
            geometry_field_format=properties.get(
                DelimitedFormat._GEOMETRY_FIELD_FORMAT_KEY
            ),
            date_format=properties.get(DelimitedFormat._DATE_FORMAT_KEY),
            record_terminator=properties.get(DelimitedFormat._RECORD_TERMINATOR_KEY),
            field_delimiter=properties.get(DelimitedFormat._FIELD_DELIMITER_KEY),
            has_header_row=properties.get(DelimitedFormat._HAS_HEADER_ROW),
        )

        return obj


@dataclass
class EsriJsonFormat(_FormatBase):
    T = TypeVar("EsriJsonFormat")
    name: ClassVar[str] = "esri-json"

    # format keys
    _DATE_FORMAT_KEY: str = field(init=False, default=f"{name}.dateFormat")

    # this format's instance properties
    date_format: Optional[str] = None

    def _build(self) -> dict:
        result_dict = {_FORMAT_NAME_KEY: self.name}
        properties_dict = {}

        if self.date_format is not None:
            properties_dict[EsriJsonFormat._DATE_FORMAT_KEY] = self.date_format

        if bool(properties_dict):
            result_dict["properties"] = properties_dict

        return result_dict

    @classmethod
    def _from_config(cls, config: dict) -> T:
        properties = config["properties"]
        obj = EsriJsonFormat(date_format=EsriJsonFormat._DATE_FORMAT_KEY)

        return obj


@dataclass
class GeoJsonFormat(_FormatBase):
    T = TypeVar("GeoJsonFormat")
    name: ClassVar[str] = "geo-json"

    # format keys
    _DATE_FORMAT_KEY: str = field(init=False, default=f"{name}.dateFormat")

    # this format's instance properties
    date_format: Optional[str] = None

    def _build(self) -> dict:
        result_dict = {_FORMAT_NAME_KEY: self.name}
        properties_dict = {}

        if self.date_format is not None:
            properties_dict[GeoJsonFormat._DATE_FORMAT_KEY] = self.date_format

        if bool(properties_dict):
            result_dict["properties"] = properties_dict

        return result_dict

    @classmethod
    def _from_config(cls, config: dict) -> T:
        properties = config["properties"]
        obj = GeoJsonFormat(date_format=GeoJsonFormat._DATE_FORMAT_KEY)

        return obj


@dataclass
class JsonFormat(_FormatBase):
    T = TypeVar("JsonFormat")
    name: ClassVar[str] = "json"

    # format keys
    _BUILD_GEOMETRY_FROM_FIELDS_KEY: str = field(
        init=False, default=f"{name}.buildGeometryFromFields"
    )
    _X_FIELD_KEY: str = field(init=False, default=f"{name}.xField")
    _Y_FIELD_KEY: str = field(init=False, default=f"{name}.yField")
    _HAS_Z_FIELD_KEY_KEY: str = field(init=False, default=f"{name}.hasZField")
    _Z_FIELD_KEY: str = field(init=False, default=f"{name}.zField")
    _Z_UNIT_KEY: str = field(init=False, default=f"{name}.zUnit")
    _GEOMETRY_FIELD_KEY: str = field(init=False, default=f"{name}.geometryField")
    _GEOMETRY_FIELD_FORMAT_KEY: str = field(
        init=False, default=f"{name}.geometryFieldFormat"
    )

    _DATE_FORMAT_KEY: str = field(init=False, default=f"{name}.dateFormat")

    _ROOT_NODE_KEY: str = field(init=False, default=f"{name}.rootNode")
    _FLATTEN_KEY: str = field(init=False, default=f"{name}.flatten")
    _FIELD_FLATTENING_EXEMPTIONS_KEY: str = field(
        init=False, default=f"{name}.fieldFlatteningExemptions"
    )

    # this format's instance properties
    root_node: Optional[str] = None
    flatten: Optional[bool] = None
    field_flattening_exemptions: Optional[str] = None
    # date properties
    date_format: Optional[str] = None
    # geometry properties
    build_geometry_from_fields: Optional[bool] = None
    x_field: Optional[str] = None
    y_field: Optional[str] = None
    has_z_field: Optional[bool] = None
    z_field: Optional[str] = None
    z_unit: Optional[int] = None
    geometry_field: Optional[str] = None
    geometry_field_format: Optional[str] = None

    def _build(self) -> dict:
        result_dict = {_FORMAT_NAME_KEY: self.name}
        properties_dict = {}

        if self.build_geometry_from_fields is not None:
            properties_dict[
                JsonFormat._BUILD_GEOMETRY_FROM_FIELDS_KEY
            ] = self.build_geometry_from_fields
        if self.x_field is not None:
            properties_dict[JsonFormat._X_FIELD_KEY] = self.x_field
        if self.y_field is not None:
            properties_dict[JsonFormat._Y_FIELD_KEY] = self.y_field
        if self.has_z_field is not None:
            properties_dict[JsonFormat._HAS_Z_FIELD_KEY_KEY] = self.has_z_field
        if self.z_field is not None:
            properties_dict[JsonFormat._Z_FIELD_KEY] = self.z_field
        if self.z_unit is not None:
            properties_dict[JsonFormat._Z_UNIT_KEY] = self.z_unit
        if self.geometry_field is not None:
            properties_dict[JsonFormat._GEOMETRY_FIELD_KEY] = self.geometry_field
        if self.geometry_field_format is not None:
            properties_dict[
                JsonFormat._GEOMETRY_FIELD_FORMAT_KEY
            ] = self.geometry_field_format
        if self.date_format is not None:
            properties_dict[JsonFormat._DATE_FORMAT_KEY] = self.date_format
        if self.root_node is not None:
            properties_dict[JsonFormat._ROOT_NODE_KEY] = self.root_node
        if self.flatten is not None:
            properties_dict[JsonFormat._FLATTEN_KEY] = self.flatten
        if self.field_flattening_exemptions is not None:
            properties_dict[
                JsonFormat._FIELD_FLATTENING_EXEMPTIONS_KEY
            ] = self.field_flattening_exemptions

        if bool(properties_dict):
            result_dict["properties"] = properties_dict

        return result_dict

    @classmethod
    def _from_config(cls, config: dict) -> T:
        properties = config["properties"]
        obj = JsonFormat(
            build_geometry_from_fields=properties.get(
                JsonFormat._BUILD_GEOMETRY_FROM_FIELDS_KEY
            ),
            x_field=properties.get(JsonFormat._X_FIELD_KEY),
            y_field=properties.get(JsonFormat._Y_FIELD_KEY),
            has_z_field=properties.get(JsonFormat._HAS_Z_FIELD_KEY_KEY),
            z_field=properties.get(JsonFormat._Z_FIELD_KEY),
            z_unit=properties.get(JsonFormat._Z_UNIT_KEY),
            geometry_field=properties.get(JsonFormat._GEOMETRY_FIELD_KEY),
            geometry_field_format=properties.get(JsonFormat._GEOMETRY_FIELD_FORMAT_KEY),
            date_format=properties.get(JsonFormat._DATE_FORMAT_KEY),
            root_node=properties.get(JsonFormat._ROOT_NODE_KEY),
            flatten=properties.get(JsonFormat._FLATTEN_KEY),
            field_flattening_exemptions=properties.get(
                JsonFormat._FIELD_FLATTENING_EXEMPTIONS_KEY
            ),
        )

        return obj


@dataclass
class XMLFormat(_FormatBase):
    T = TypeVar("XMLFormat")
    name: ClassVar[str] = "xml"

    # format keys
    _BUILD_GEOMETRY_FROM_FIELDS_KEY: str = field(
        init=False, default=f"{name}.buildGeometryFromFields"
    )
    _X_FIELD_KEY: str = field(init=False, default=f"{name}.xField")
    _Y_FIELD_KEY: str = field(init=False, default=f"{name}.yField")
    _HAS_Z_FIELD_KEY_KEY: str = field(init=False, default=f"{name}.hasZField")
    _Z_FIELD_KEY: str = field(init=False, default=f"{name}.zField")
    _Z_UNIT_KEY: str = field(init=False, default=f"{name}.zUnit")
    _GEOMETRY_FIELD_KEY: str = field(init=False, default=f"{name}.geometryField")
    _GEOMETRY_FIELD_FORMAT_KEY: str = field(
        init=False, default=f"{name}.geometryFieldFormat"
    )

    _DATE_FORMAT_KEY: str = field(init=False, default=f"{name}.dateFormat")

    _ROOT_NODE_KEY: str = field(init=False, default=f"{name}.rootNode")
    _FLATTEN_KEY: str = field(init=False, default=f"{name}.flatten")
    _FIELD_FLATTENING_EXEMPTIONS_KEY: str = field(
        init=False, default=f"{name}.fieldFlatteningExemptions"
    )

    # this format's instance properties
    root_node: Optional[str] = None
    flatten: Optional[bool] = None
    field_flattening_exemptions: Optional[str] = None
    # date properties
    date_format: Optional[str] = None
    # geometry properties
    build_geometry_from_fields: Optional[bool] = None
    x_field: Optional[str] = None
    y_field: Optional[str] = None
    has_z_field: Optional[bool] = None
    z_field: Optional[str] = None
    z_unit: Optional[int] = None
    geometry_field: Optional[str] = None
    geometry_field_format: Optional[str] = None

    def _build(self) -> dict:
        result_dict = {_FORMAT_NAME_KEY: self.name}
        properties_dict = {}

        if self.build_geometry_from_fields is not None:
            properties_dict[
                XMLFormat._BUILD_GEOMETRY_FROM_FIELDS_KEY
            ] = self.build_geometry_from_fields
        if self.x_field is not None:
            properties_dict[XMLFormat._X_FIELD_KEY] = self.x_field
        if self.y_field is not None:
            properties_dict[XMLFormat._Y_FIELD_KEY] = self.y_field
        if self.has_z_field is not None:
            properties_dict[XMLFormat._HAS_Z_FIELD_KEY_KEY] = self.has_z_field
        if self.z_field is not None:
            properties_dict[XMLFormat._Z_FIELD_KEY] = self.z_field
        if self.z_unit is not None:
            properties_dict[XMLFormat._Z_UNIT_KEY] = self.z_unit
        if self.geometry_field is not None:
            properties_dict[XMLFormat._GEOMETRY_FIELD_KEY] = self.geometry_field
        if self.geometry_field_format is not None:
            properties_dict[
                XMLFormat._GEOMETRY_FIELD_FORMAT_KEY
            ] = self.geometry_field_format
        if self.date_format is not None:
            properties_dict[XMLFormat._DATE_FORMAT_KEY] = self.date_format
        if self.root_node is not None:
            properties_dict[XMLFormat._ROOT_NODE_KEY] = self.root_node
        if self.flatten is not None:
            properties_dict[XMLFormat._FLATTEN_KEY] = self.flatten
        if self.field_flattening_exemptions is not None:
            properties_dict[
                XMLFormat._FIELD_FLATTENING_EXEMPTIONS_KEY
            ] = self.field_flattening_exemptions

        if bool(properties_dict):
            result_dict["properties"] = properties_dict

        return result_dict

    @classmethod
    def _from_config(cls, config: dict) -> T:
        properties = config["properties"]
        obj = XMLFormat(
            build_geometry_from_fields=properties.get(
                XMLFormat._BUILD_GEOMETRY_FROM_FIELDS_KEY
            ),
            x_field=properties.get(XMLFormat._X_FIELD_KEY),
            y_field=properties.get(XMLFormat._Y_FIELD_KEY),
            has_z_field=properties.get(XMLFormat._HAS_Z_FIELD_KEY_KEY),
            z_field=properties.get(XMLFormat._Z_FIELD_KEY),
            z_unit=properties.get(XMLFormat._Z_UNIT_KEY),
            geometry_field=properties.get(XMLFormat._GEOMETRY_FIELD_KEY),
            geometry_field_format=properties.get(XMLFormat._GEOMETRY_FIELD_FORMAT_KEY),
            date_format=properties.get(XMLFormat._DATE_FORMAT_KEY),
            root_node=properties.get(XMLFormat._ROOT_NODE_KEY),
            flatten=properties.get(XMLFormat._FLATTEN_KEY),
            field_flattening_exemptions=properties.get(
                XMLFormat._FIELD_FLATTENING_EXEMPTIONS_KEY
            ),
        )

        return obj


def _format_from_config(
    config: dict,
) -> Optional[
    Union[
        GeoRssFormat,
        RssFormat,
        DelimitedFormat,
        EsriJsonFormat,
        GeoJsonFormat,
        JsonFormat,
        XMLFormat,
    ]
]:
    """
    Identifies and instantiates a format object from a feed configuration json dict

    ==================     ====================================================================
    **Parameter**           **Description**
    ------------------     --------------------------------------------------------------------
    config                 dict. The feed configuration json/dict.
    ==================     ====================================================================
    """
    if _FORMAT_NAME_KEY in config:
        if config[_FORMAT_NAME_KEY] == GeoRssFormat.name:
            return GeoRssFormat._from_config(config)
        elif config[_FORMAT_NAME_KEY] == RssFormat.name:
            return RssFormat._from_config(config)
        elif config[_FORMAT_NAME_KEY] == DelimitedFormat.name:
            return DelimitedFormat._from_config(config)
        elif config[_FORMAT_NAME_KEY] == EsriJsonFormat.name:
            return EsriJsonFormat._from_config(config)
        elif config[_FORMAT_NAME_KEY] == GeoJsonFormat.name:
            return GeoJsonFormat._from_config(config)
        elif config[_FORMAT_NAME_KEY] == JsonFormat.name:
            return JsonFormat._from_config(config)
        elif config[_FORMAT_NAME_KEY] == XMLFormat.name:
            return XMLFormat._from_config(config)
        else:
            return None
    else:
        return None



from __future__ import annotations
import json
import logging
import datetime

from arcgis.gis import Layer, _GISResource

# Supported Data Types
from typing import Any, Optional, Union
from arcgis.geometry import Point, Polygon, Polyline
from arcgis.features import Feature, FeatureSet
from arcgis.features import FeatureLayer, Table
from arcgis.network import _utils
from arcgis._impl.common._utils import _validate_url

try:
    import pandas as pd
    from arcgis.features.geo import _is_geoenabled

    HASPANDAS = True
except ImportError:
    HASPANDAS = False

    def _is_geoenabled(df):
        return False


from arcgis.gis import Item

_log = logging.getLogger(__name__)


###########################################################################
def _handle_spatial_inputs(data, do_not_locate=True, has_z=False, where=None):
    """
    Handles the various supported inputs types
    """
    template = {
        "type": "features",
        "doNotLocateOnRestrictedElements": do_not_locate,
        "hasZ": has_z,
    }
    if isinstance(data, Item) and data.type in ["Feature Layer", "Feature Service"]:
        return _handle_spatial_inputs(data.layers[0])
    if HASPANDAS and isinstance(data, pd.DataFrame) and _is_geoenabled(df=data):
        return data.spatial.__feature_set__
    elif isinstance(data, FeatureSet):
        return data.sdf.spatial.__feature_set__
    elif isinstance(data, list):
        stops_dict = []
        for stop in data:
            if isinstance(stop, Feature):
                stops_dict.append(stop.as_dict)
            else:
                stops_dict.append(stop)
        template["features"] = stops_dict
        return template
    elif isinstance(data, (FeatureLayer, Table)):
        from urllib.parse import quote
        import json

        query = data.filter
        url = data._url
        if query and len(str(query)) > 0:
            query = quote(query)
            url += "/query?where=%s&outFields=*&f=json" % query
            if data._gis._con.token:
                url += "&token=%s" % data._gis._con.token
        else:
            query = quote("1=1")
            url += "/query?where=%s&outFields=*&f=json" % query
            if data._gis._con.token:
                url += "&token=%s" % data._gis._con.token
        template["url"] = url
        return template
    else:
        return data
    return data


###########################################################################
class NAJob(object):
    """Represents a Future Job for Network Analyst Jobs"""

    _future = None
    _gis = None
    _start_time = None
    _end_time = None

    # ----------------------------------------------------------------------
    def __init__(self, future, task=None, notify=False):
        """
        initializer
        """
        if task is None:
            self._task = "Network Task"
        else:
            self._task = task
        self._future = future
        self._start_time = datetime.datetime.now()
        if notify:
            self._future.add_done_callback(self._notify)
        self._future.add_done_callback(self._set_end_time)

    # ----------------------------------------------------------------------
    @property
    def ellapse_time(self):
        """
        Returns the Ellapse Time for the Job
        """
        if self._end_time:
            return self._end_time - self._start_time
        else:
            return datetime.datetime.now() - self._start_time

    # ----------------------------------------------------------------------
    def _set_end_time(self, future):
        """sets the finish time"""
        self._end_time = datetime.datetime.now()

    # ----------------------------------------------------------------------
    def _notify(self, future):
        """prints finished method"""
        jobid = str(self).replace("<", "").replace(">", "")
        try:
            res = future.result()
            infomsg = "{jobid} finished successfully.".format(jobid=jobid)
            _log.info(infomsg)
            print(infomsg)
        except Exception as e:
            msg = str(e)
            msg = "{jobid} failed: {msg}".format(jobid=jobid, msg=msg)
            _log.info(msg)
            print(msg)

    # ----------------------------------------------------------------------
    def __str__(self):
        task = self._task
        return f"<{task} Job>"

    # ----------------------------------------------------------------------
    def __repr__(self):
        return self.__str__()

    # ----------------------------------------------------------------------
    @property
    def status(self):
        """
        returns the GP status

        :return: String
        """
        return self._future.done()

    # ----------------------------------------------------------------------
    def cancel(self):
        """
        Attempt to cancel the call. If the call is currently being executed
        or finished running and cannot be cancelled then the method will
        return False, otherwise the call will be cancelled and the method
        will return True.

        :return: boolean
        """

        if self.done():
            return False
        if self.cancelled():
            return False
        return True

    # ----------------------------------------------------------------------
    def cancelled(self):
        """
        Return True if the call was successfully cancelled.

        :return: boolean
        """
        return self._future.cancelled()

    # ----------------------------------------------------------------------
    def running(self):
        """
        Return True if the call is currently being executed and cannot be cancelled.

        :return: boolean
        """
        return self._future.running()

    # ----------------------------------------------------------------------
    def done(self):
        """
        Return True if the call was successfully cancelled or finished running.

        :return: boolean
        """
        return self._future.done()

    # ----------------------------------------------------------------------
    def result(self):
        """
        Return the value returned by the call. If the call hasn't yet completed
        then this method will wait.

        :return: object
        """
        if self.cancelled():
            return None
        return self._future.result()


###########################################################################
class NetworkLayer(Layer):
    """
    NetworkLayer represents a single network layer. It provides basic
    information about the network layer such as its name, type, and network
    classes. Additionally, depending on the layer type, it provides different
    pieces of information.

    It is a base class for RouteLayer, ServiceAreaLayer, and
    ClosestFacilityLayer.
    """

    # ----------------------------------------------------------------------
    def _run_async(self, fn, **inputs):
        """runs the inputs asynchronously"""
        import concurrent.futures

        tp = concurrent.futures.ThreadPoolExecutor(1)
        try:
            future = tp.submit(fn=fn, **inputs)
        except:
            future = tp.submit(fn, **inputs)
        tp.shutdown(False)
        return future

    # ----------------------------------------------------------------------
    def retrieve_travel_modes(self):
        """identify all the valid travel modes that have been defined on the
        network dataset or in the portal if the GIS server is federated"""
        url = self._url + "/retrieveTravelModes"
        params = {"f": "json"}
        return self._con.get(
            path=url,
            params=params,
        )


###########################################################################
class RouteLayer(NetworkLayer):
    """
    The Route Layer which has common properties of Network Layer
    as well as some attributes unique to Route Network Layer only.
    """

    def solve(
        self,
        stops: Union[list[Point], list[FeatureSet], FeatureSet],
        barriers: Optional[Union[Point, FeatureSet, dict[str, Any]]] = None,
        polyline_barriers: Optional[Union[Polyline, FeatureSet, dict[str, Any]]] = None,
        polygon_barriers: Optional[Union[Polygon, FeatureSet, dict[str, Any]]] = None,
        travel_mode: Optional[str] = None,
        attribute_parameter_values: Optional[Union[str, list[str]]] = None,
        return_directions: bool = True,
        return_routes: bool = True,
        return_stops: bool = False,
        return_barriers: bool = False,
        return_polyline_barriers: bool = True,
        return_polygon_barriers: bool = True,
        out_sr: Optional[int] = None,
        ignore_invalid_locations: bool = True,
        output_lines: Optional[str] = None,
        find_best_sequence: bool = False,
        preserve_first_stop: bool = True,
        preserve_last_stop: bool = True,
        use_time_windows: bool = False,
        start_time: Optional[str] = None,
        start_time_is_utc: bool = False,
        accumulate_attribute_names: Optional[str] = None,
        impedance_attribute_name: Optional[str] = None,
        restriction_attribute_names: Optional[str] = None,
        restrict_u_turns: Optional[bool] = None,
        use_hierarchy: bool = True,
        directions_language: Optional[str] = None,
        directions_output_type: Optional[str] = None,
        directions_style_name: Optional[str] = None,
        directions_length_units: Optional[str] = None,
        directions_time_attribute_name: Optional[str] = None,
        output_geometry_precision: Optional[float] = None,
        output_geometry_precision_units: Optional[str] = None,
        return_z: bool = False,
        overrides: Optional[dict[str, Any]] = None,
        preserve_objectid: bool = False,
        future: bool = False,
        time_windows_are_utc: bool = False,
        return_traversed_edges: Optional[bool] = None,
        return_traversed_junctions: Optional[bool] = None,
        return_traversed_turns: Optional[bool] = None,
        geometry_precision: Optional[int] = None,
        geometry_precision_z: Optional[int] = None,
        geometry_precision_m: Optional[int] = None,
    ):
        """
        The solve operation is performed on a network layer resource.
        The solve operation is supported on a network layer whose layerType
        is esriNAServerRouteLayer. You can provide arguments to the solve
        route operation as query parameters.


        ===================================     ====================================================================
        **Parameter**                            **Description**
        -----------------------------------     --------------------------------------------------------------------
        stops                                   Required Points/FeatureSet/a list of Features. The set of stops
                                                loaded as network locations during analysis. Stops can be specified
                                                using a simple comma / semi-colon based syntax or as a JSON
                                                structure. If stops are not specified, preloaded stops from the map
                                                document are used in the analysis.
        -----------------------------------     --------------------------------------------------------------------
        barriers                                Optional Point/FeatureSet. The set of barriers loaded as network
                                                locations during analysis. Barriers can be specified using a simple
                                                comma/semi-colon based syntax or as a JSON structure. If barriers
                                                are not specified, preloaded barriers from the map document are used
                                                in the analysis. If an empty json object is passed ('{}') preloaded
                                                barriers are ignored.
        -----------------------------------     --------------------------------------------------------------------
        polyline_barriers                       Optional Polyline/FeatureSet. The set of polyline barriers loaded
                                                as network locations during analysis. If polyline barriers are not
                                                specified, preloaded polyline barriers from the map document are
                                                used in the analysis. If an empty json object is passed ('{}')
                                                preloaded polyline barriers are ignored.
        -----------------------------------     --------------------------------------------------------------------
        polygon_barriers                        Optional Polygon/FeatureSet. The set of polygon barriers loaded as
                                                network locations during analysis. If polygon barriers are not
                                                specified, preloaded polygon barriers from the map document are used
                                                in the analysis. If an empty json object is passed ('{}') preloaded
                                                polygon barriers are ignored.
        -----------------------------------     --------------------------------------------------------------------
        travel_mode                             Optional string. Travel modes provide override values that help you
                                                quickly and consistently model a vehicle or mode of transportation.
                                                The chosen travel mode must be preconfigured on the network dataset
                                                that the routing service references.
        -----------------------------------     --------------------------------------------------------------------
        attribute_parameter_values              Optional string/list.  A set of attribute parameter values that can be
                                                parameterized to determine which network elements can be used by a
                                                vehicle.
        -----------------------------------     --------------------------------------------------------------------
        return_directions                       Optional boolean. If true, directions will be generated and returned
                                                with the analysis results. Default is true.
        -----------------------------------     --------------------------------------------------------------------
        return_routes                           Optional boolean. If true, routes will be returned with the analysis
                                                results. Default is true.
        -----------------------------------     --------------------------------------------------------------------
        return_stops                            Optional boolean.  If true, stops will be returned with the analysis
                                                results. Default is false.
        -----------------------------------     --------------------------------------------------------------------
        return_barriers                         Optional boolean.  If true, barriers will be returned with the
                                                analysis results. Default is false.
        -----------------------------------     --------------------------------------------------------------------
        return_polyline_barriers                Optional boolean. If true, polyline barriers will be returned with
                                                the analysis results. Default is false.
        -----------------------------------     --------------------------------------------------------------------
        return_polygon_barriers                 Optional boolean. If true, polygon barriers will be returned with
                                                the analysis results. Default is false.
        -----------------------------------     --------------------------------------------------------------------
        out_sr                                  Optional Integer. The spatial reference of the geometries returned
                                                with the analysis results.
        -----------------------------------     --------------------------------------------------------------------
        ignore_invalid_locations                Optional boolean. - If true, the solver will ignore invalid
                                                locations. Otherwise, it will raise an error. Default is true.
        -----------------------------------     --------------------------------------------------------------------
        output_lines                            The type of output lines to be generated in the result. The default
                                                is as defined in the network layer.
                                                Values: esriNAOutputLineTrueShape |
                                                        esriNAOutputLineTrueShapeWithMeasure |
                                                        esriNAOutputLineStraight | esriNAOutputLineNone
        -----------------------------------     --------------------------------------------------------------------
        find_best_sequence                      Optional boolean. If true, the solver should re-sequence the route in
                                                the optimal order. The default is as defined in the network layer.
        -----------------------------------     --------------------------------------------------------------------
        preserve_first_stop                     Optional boolean. If true, the solver should keep the first stop
                                                fixed in the sequence. The default is as defined in the network
                                                layer.
        -----------------------------------     --------------------------------------------------------------------
        preserve_last_stop                      Optional boolean. If true, the solver should keep the last stop fixed
                                                in the sequence. The default is as defined in the network layer.
        -----------------------------------     --------------------------------------------------------------------
        use_time_window                         Optional boolean. If true, the solver should consider time windows.
                                                The default is as defined in the network layer.
        -----------------------------------     --------------------------------------------------------------------
        start_time                              Optional string. The time the route begins. If not specified, the
                                                solver will use the default as defined in the network layer.
        -----------------------------------     --------------------------------------------------------------------
        start_time_is_utc                       Optional boolean. The time zone of the startTime parameter.
        -----------------------------------     --------------------------------------------------------------------
        accumulate_attribute_names              Optional string. A list of network attribute names to be accumulated
                                                with the analysis. The default is as defined in the network layer.
                                                The value should be specified as a comma separated list of attribute
                                                names. You can also specify a value of none to indicate that no
                                                network attributes should be accumulated.
        -----------------------------------     --------------------------------------------------------------------
        impedance_attribute_name                Optional string. The network attribute name to be used as the impedance
                                                attribute in analysis. The default is as defined in the network layer.
        -----------------------------------     --------------------------------------------------------------------
        restriction_attribute_names             Optional string. -The list of network attribute names to be
                                                used as restrictions with the analysis. The default is as defined in
                                                the network layer. The value should be specified as a comma
                                                separated list of attribute names. You can also specify a value of
                                                none to indicate that no network attributes should be used as
                                                restrictions.
        -----------------------------------     --------------------------------------------------------------------
        restrict_u_turns                        Optional boolean. Specifies how U-Turns should be restricted in the
                                                analysis. The default is as defined in the network layer.
                                                Values: esriNFSBAllowBacktrack | esriNFSBAtDeadEndsOnly |
                                                        esriNFSBNoBacktrack | esriNFSBAtDeadEndsAndIntersections
        -----------------------------------     --------------------------------------------------------------------
        use_hierarchy                           Optional boolean.  If true, the hierarchy attribute for the network
                                                should be used in analysis. The default is as defined in the network
                                                layer.
        -----------------------------------     --------------------------------------------------------------------
        directions_language                     Optional string. The language to be used when computing directions.
                                                The default is the language of the server's operating system. The list
                                                of supported languages can be found in REST layer description.
        -----------------------------------     --------------------------------------------------------------------
        directions_output_type                  Optional string.  Defines content, verbosity of returned directions.
                                                The default is esriDOTInstructionsOnly.
                                                Values: esriDOTComplete | esriDOTCompleteNoEvents
                                                        | esriDOTInstructionsOnly | esriDOTStandard |
                                                        esriDOTSummaryOnly
        -----------------------------------     --------------------------------------------------------------------
        directions_style_name                   Optional string. The style to be used when returning the directions.
                                                The default is as defined in the network layer. The list of
                                                supported styles can be found in REST layer description.
        -----------------------------------     --------------------------------------------------------------------
        directions_length_units                 Optional string. The length units to use when computing directions.
                                                The default is as defined in the network layer.
                                                Values: esriNAUFeet | esriNAUKilometers | esriNAUMeters |
                                                        esriNAUMiles | esriNAUNauticalMiles | esriNAUYards |
                                                        esriNAUUnknown
        -----------------------------------     --------------------------------------------------------------------
        directions_time_attribute_name          Optional string. The name of network attribute to use for the drive
                                                time when computing directions. The default is as defined in the network
                                                layer.
        -----------------------------------     --------------------------------------------------------------------
        output_geometry_precision               Optional float.  The precision of the output geometry after
                                                generalization. If 0, no generalization of output geometry is
                                                performed. The default is as defined in the network service
                                                configuration.
        -----------------------------------     --------------------------------------------------------------------
        output_geometry_precision_units         Optional string. The units of the output geometry precision. The
                                                default value is esriUnknownUnits.
                                                Values: esriUnknownUnits | esriCentimeters | esriDecimalDegrees |
                                                        esriDecimeters | esriFeet | esriInches | esriKilometers |
                                                        esriMeters | esriMiles | esriMillimeters |
                                                        esriNauticalMiles | esriPoints | esriYards
        -----------------------------------     --------------------------------------------------------------------
        return_z                                Optional boolean. If true, Z values will be included in the returned
                                                routes and compressed geometry if the network dataset is Z-aware.
                                                The default is false.
        -----------------------------------     --------------------------------------------------------------------
        overrides                               Optional dictionary. Specify additional settings that can influence
                                                the behavior of the solver.  A list of supported override settings
                                                for each solver and their acceptable values can be obtained by
                                                contacting Esri Technical Support.
        -----------------------------------     --------------------------------------------------------------------
        preserve_objectid                       Optional Boolean.  If True, all objectid values are maintained.  The
                                                default is False.
        -----------------------------------     --------------------------------------------------------------------
        future                                  Optional boolean. If True, a future object will be returned and the process
                                                will not wait for the task to complete. The default is False, which means wait for results.
        -----------------------------------     --------------------------------------------------------------------
        time_windows_are_utc                    Optional boolean. Specify whether the TimeWindowStart and TimeWindowEnd
                                                attribute values on stops are specified in coordinated universal time (UTC)
                                                or geographically local time.
        -----------------------------------     --------------------------------------------------------------------
        return_traversed_edges                  Optional boolean. Specify whether traversed edges will be returned by the service.
        -----------------------------------     --------------------------------------------------------------------
        return_traversed_junctions              Optional boolean. Specify whether traversed junctions will be returned by the service.
        -----------------------------------     --------------------------------------------------------------------
        return_traversed_turns                  Optional boolean. Specify whether traversed turns will be returned by the service.
        -----------------------------------     --------------------------------------------------------------------
        geometry_precision                      Optional Integer. Use this parameter to specify the number of decimal places in the response geometries returned by solve operation. This applies to x/y values only (not m- or z-values).
        -----------------------------------     --------------------------------------------------------------------
        geometry_precision_z                    Optional Integer. Use this parameter specify the number of decimal places in the response geometries returned by solve operation. This applies to z-value only.
        -----------------------------------     --------------------------------------------------------------------
        geometry_precision_m                    Optional Integer. Use this parameter to specify the number of decimal places in the response geometries returned by solve operation. This applies to m-value only.
        ===================================     ====================================================================


        :return: dict

        .. code-block:: python

            # USAGE EXAMPLE: Solving the routing problem by passing in a FeatureSet

            # get a FeatureSet through query
            fl = sample_cities.layers[0]
            cities_to_visit = fl.query(where="ST = 'CA' AND POP2010 > 300000",
                                       out_fields='NAME', out_sr=4326)

            type(cities_to_visit)
            >> arcgis.features.feature.FeatureSet

            # pass in the FeatureSet
            result = route_layer.solve(stops=cities_to_visit, preserve_first_stop=True,
                                       preserve_last_stop=True, find_best_sequence=True, return_directions=False,
                                       return_stops=True, return_barriers=False, return_polygon_barriers=False,
                                       return_polyline_barriers=False, return_routes=True,
                                       output_lines='esriNAOutputLineStraight')

        """

        if not self.properties.layerType == "esriNAServerRouteLayer":
            raise ValueError(
                "The solve operation is supported on a network "
                "layer of Route type only"
            )

        url = self._url + "/solve"

        params = {
            "f": "json",
        }
        if travel_mode:
            if isinstance(travel_mode, str):
                travel_mode = _utils.find_travel_mode(
                    gis=self._gis, travel_mode=travel_mode
                )
                params["travel_mode"] = travel_mode
            elif isinstance(travel_mode, dict):
                travel_mode = json.dumps(travel_mode)
                params["travel_mode"] = travel_mode
            else:
                travel_mode = _utils.find_travel_mode(
                    gis=self._gis, travel_mode=_utils.default_travel_mode(gis=self._gis)
                )
                params["travel_mode"] = travel_mode

        stops = _handle_spatial_inputs(data=stops)
        params["stops"] = stops
        if directions_output_type is None:
            directions_output_type = "esriDOTInstructionsOnly"
        if not barriers is None:
            params["barriers"] = _handle_spatial_inputs(data=barriers)
        if not polyline_barriers is None:
            params["polylineBarriers"] = _handle_spatial_inputs(data=polyline_barriers)
        if not polygon_barriers is None:
            params["polygonBarriers"] = _handle_spatial_inputs(data=polygon_barriers)
        if not travel_mode is None:
            params["travelMode"] = travel_mode
        if not attribute_parameter_values is None:
            params["attributeParameterValues"] = attribute_parameter_values
        if not return_directions is None:
            params["returnDirections"] = return_directions
        if not return_routes is None:
            params["returnRoutes"] = return_routes
        if not return_stops is None:
            params["returnStops"] = return_stops
        if not return_barriers is None:
            params["returnBarriers"] = return_barriers
        if not return_polyline_barriers is None:
            params["returnPolylineBarriers"] = return_polyline_barriers
        if not return_polygon_barriers is None:
            params["returnPolygonBarriers"] = return_polygon_barriers
        if not out_sr is None:
            params["outSR"] = out_sr
        if not ignore_invalid_locations is None:
            params["ignoreInvalidLocations"] = ignore_invalid_locations
        if not output_lines is None:
            params["outputLines"] = output_lines
        if not find_best_sequence is None:
            params["findBestSequence"] = find_best_sequence
        if not preserve_first_stop is None:
            params["preserveFirstStop"] = preserve_first_stop
        if not preserve_last_stop is None:
            params["preserveLastStop"] = preserve_last_stop
        if not use_time_windows is None:
            params["useTimeWindows"] = use_time_windows
        if not time_windows_are_utc is None:
            params["timeWindowsAreUTC"] = time_windows_are_utc
        if not start_time is None:
            if isinstance(start_time, datetime.datetime):
                start_time = f"{start_time.timestamp() * 1000}"
            params["startTime"] = start_time
        if not start_time_is_utc is None:
            params["startTimeIsUTC"] = start_time_is_utc
        if not accumulate_attribute_names is None:
            params["accumulateAttributeNames"] = accumulate_attribute_names
        if not impedance_attribute_name is None:
            params["impedanceAttributeName"] = impedance_attribute_name
        if not restriction_attribute_names is None:
            params["restrictionAttributeNames"] = restriction_attribute_names
        if not restrict_u_turns is None:
            params["restrictUTurns"] = restrict_u_turns
        if not use_hierarchy is None:
            params["useHierarchy"] = use_hierarchy
        if not directions_language is None:
            params["directionsLanguage"] = directions_language
        if not directions_output_type is None:
            params["directionsOutputType"] = directions_output_type
        if not directions_style_name is None:
            params["directionsStyleName"] = directions_style_name
        if not directions_length_units is None:
            params["directionsLengthUnits"] = directions_length_units
        if not directions_time_attribute_name is None:
            params["directionsTimeAttributeName"] = directions_time_attribute_name
        if not output_geometry_precision is None:
            params["outputGeometryPrecision"] = output_geometry_precision
        if not output_geometry_precision_units is None:
            params["outputGeometryPrecisionUnits"] = output_geometry_precision_units
        if not return_z is None:
            params["returnZ"] = return_z
        if not overrides is None:
            params["overrides"] = overrides
        if not preserve_objectid is None:
            params["preserveObjectID"] = preserve_objectid
        if not geometry_precision is None:
            params["geometryPrecision"] = geometry_precision
        if not geometry_precision_z is None:
            params["geometryPrecisionZ"] = geometry_precision_z
        if not geometry_precision_m is None:
            params["geometryPrecisionM"] = geometry_precision_m
        if not return_traversed_edges is None:
            params["returnTraversedEdges"] = return_traversed_edges
        if not return_traversed_junctions is None:
            params["returnTraversedJunctions"] = return_traversed_junctions
        if not return_traversed_turns is None:
            params["returnTraversedTurns"] = return_traversed_turns

        if future:
            f = self._run_async(
                self._con.post,
                **{"path": url, "postdata": params},
            )
            return NAJob(future=f, task="RouteLayer Solve")
        return self._con.post(path=url, postdata=params)  # ,


###########################################################################
class ServiceAreaLayer(NetworkLayer):
    """
    The Service Area Layer which has common properties of Network
    Layer as well as some attributes unique to Service Area Layer
    only.
    """

    def solve_service_area(
        self,
        facilities: Union[FeatureSet, Point, list, dict],
        barriers: Optional[Union[Point, FeatureSet, dict[str, Any]]] = None,
        polyline_barriers: Optional[Union[Polyline, FeatureSet, dict[str, Any]]] = None,
        polygon_barriers: Optional[Union[Polygon, FeatureSet, dict[str, Any]]] = None,
        travel_mode: Optional[str] = None,
        attribute_parameter_values: Optional[Union[str, list]] = None,
        default_breaks: Optional[list[float]] = None,
        exclude_sources_from_polygons: Optional[list[str]] = None,
        merge_similar_polygon_ranges: Optional[bool] = None,
        output_lines: Optional[str] = None,
        output_polygons: Optional[str] = None,
        overlap_lines: Optional[str] = None,
        overlap_polygons: Optional[str] = None,
        split_lines_at_breaks: Optional[bool] = None,
        split_polygons_at_breaks: Optional[bool] = None,
        trim_outer_polygon: Optional[bool] = None,
        trim_polygon_distance: Optional[Union[str, int]] = None,
        trim_polygon_distance_units: Optional[str] = None,
        return_facilities: bool = False,
        return_barriers: bool = False,
        return_polyline_barriers: bool = False,
        return_polygon_barriers: bool = False,
        out_sr: Optional[int] = None,
        accumulate_attribute_names: Optional[list[str]] = None,
        impedance_attribute_name: Optional[str] = None,
        restriction_attribute_names: Optional[list[str]] = None,
        restrict_u_turns: Optional[str] = None,
        output_geometry_precision: Optional[int] = None,
        output_geometry_precision_units: str = "esriUnknownUnits",
        use_hierarchy: Optional[bool] = None,
        time_of_day: Optional[datetime.datetime] = None,
        time_of_day_is_utc: Optional[bool] = None,
        travel_direction: Optional[str] = None,
        return_z: bool = False,
        overrides: Optional[dict[str, Any]] = None,
        preserve_objectid: bool = False,
        future: bool = False,
        ignore_invalid_locations: bool = True,
        geometry_precision: Optional[int] = None,
        geometry_precision_z: Optional[int] = None,
        geometry_precision_m: Optional[int] = None,
    ):
        """The solve service area operation is performed on a network layer
        resource of type service area (layerType is esriNAServerServiceArea).
        You can provide arguments to the solve service area operation as
        query parameters.

        ===================================     ====================================================================
        **Parameter**                            **Description**
        -----------------------------------     --------------------------------------------------------------------
        facilities                              The set of facilities loaded as network locations
                                                during analysis. Facilities can be specified using
                                                a simple comma / semi-colon based syntax or as a
                                                JSON structure. If facilities are not specified,
                                                preloaded facilities from the map document are used
                                                in the analysis. If an empty json object is passed
                                                ('{}') preloaded facilities are ignored.
        -----------------------------------     --------------------------------------------------------------------
        barriers                                The set of barriers loaded as network locations during
                                                analysis. Barriers can be specified using a simple
                                                comma/semicolon-based syntax or as a JSON structure.
                                                If barriers are not specified, preloaded barriers from
                                                the map document are used in the analysis. If an empty
                                                json object is passed ('{}'), preloaded barriers are
                                                ignored.
        -----------------------------------     --------------------------------------------------------------------
        polyline_barriers                       The set of polyline barriers loaded as network
                                                locations during analysis. If polyline barriers
                                                are not specified, preloaded polyline barriers
                                                from the map document are used in the analysis.
                                                If an empty json object is passed ('{}'),
                                                preloaded polyline barriers are ignored.
        -----------------------------------     --------------------------------------------------------------------
        polygon_barriers                        The set of polygon barriers loaded as network
                                                locations during analysis. If polygon barriers
                                                are not specified, preloaded polygon barriers
                                                from the map document are used in the analysis.
                                                If an empty json object is passed ('{}'),
                                                preloaded polygon barriers are ignored.
        -----------------------------------     --------------------------------------------------------------------
        travel_mode                             Travel modes provide override values that help you
                                                quickly and consistently model a vehicle or mode of
                                                transportation. The chosen travel mode must be
                                                preconfigured on the network dataset that the
                                                service area service references.
        -----------------------------------     --------------------------------------------------------------------
        attribute_parameter_values              A set of attribute parameter values that
                                                can be parameterized to determine which
                                                network elements can be used by a vehicle.
        -----------------------------------     --------------------------------------------------------------------
        default_breaks                          A comma-separated list of doubles. The default is
                                                defined in the network analysis layer.
        -----------------------------------     --------------------------------------------------------------------
        exclude_sources_from_polygons           A comma-separated list of string names.
                                                The default is defined in the network analysis layer.
        -----------------------------------     --------------------------------------------------------------------
        merge_similar_polygon_ranges            If true, similar ranges will be merged in the result polygons.
                                                The default is defined in the network analysis layer.
        -----------------------------------     --------------------------------------------------------------------
        output_lines                            The type of lines(s) generated. The default is as
                                                defined in the network analysis layer.
                                                Values: esriNAOutputLineNone | esriNAOutputLineTrueShape |
                                                esriNAOutputLineTrueShapeWithMeasure
        -----------------------------------     --------------------------------------------------------------------
        output_polygons                         The type of polygon(s) generated. The default is
                                                as defined in the network analysis layer.
        -----------------------------------     --------------------------------------------------------------------
        overlap_lines                           Indicates if the lines should overlap from multiple
                                                facilities. The default is defined in the network
                                                analysis layer.
        -----------------------------------     --------------------------------------------------------------------
        overlap_polygons                        Indicates if the polygons for all facilities
                                                should overlap. The default is defined in the
                                                network analysis layer.
        -----------------------------------     --------------------------------------------------------------------
        splitLines_at_breaks                    If true, lines will be split at breaks. The
                                                default is defined in the network analysis
                                                layer.
        -----------------------------------     --------------------------------------------------------------------
        split_polygons_at_breaks                If true, polygons will be split at breaks.
                                                The default is defined in the network
                                                analysis layer.
        -----------------------------------     --------------------------------------------------------------------
        trim_outer_polygon                      If true, the outermost polygon (at the maximum
                                                break value) will be trimmed. The default is
                                                defined in the network analysis layer.
        -----------------------------------     --------------------------------------------------------------------
        trim_polygon_distance                   If polygons are being trimmed, provides the
                                                distance to trim. The default is defined in
                                                the network analysis layer.
        -----------------------------------     --------------------------------------------------------------------
        trim_polygon_distance_units             If polygons are being trimmed, specifies
                                                the units of the trimPolygonDistance. The
                                                default is defined in the network analysis
                                                layer.
        -----------------------------------     --------------------------------------------------------------------
        return_facilities                       If true, facilities will be returned with the
                                                analysis results. Default is false.
        -----------------------------------     --------------------------------------------------------------------
        return_barriers                         If true, barriers will be returned with the analysis
                                                results. Default is false.
        -----------------------------------     --------------------------------------------------------------------
        return_polyline_barriers                If true, polyline barriers will be returned
                                                with the analysis results. Default is false.
        -----------------------------------     --------------------------------------------------------------------
        return_polygon_barriers                 If true, polygon barriers will be returned
                                                with the analysis results. Default is false.
        -----------------------------------     --------------------------------------------------------------------
        out_sr                                  The well-known ID of the spatial reference for the geometries
                                                returned with the analysis results. If outSR is not specified,
                                                the geometries are returned in the spatial reference of the map.
        -----------------------------------     --------------------------------------------------------------------
        accumulate_attribute_names              The list of network attribute names to be
                                                accumulated with the analysis. The default
                                                is as defined in the network analysis layer.
                                                The value should be specified as a comma
                                                separated list of attribute names. You can
                                                also specify a value of none to indicate that
                                                no network attributes should be accumulated.
        -----------------------------------     --------------------------------------------------------------------
        impedance_attribute_name                The network attribute name to be used as the
                                                impedance attribute in analysis. The default
                                                is as defined in the network analysis layer.
        -----------------------------------     --------------------------------------------------------------------
        restriction_attribute_names             The list of network attribute names to be
                                                used as restrictions with the analysis. The
                                                default is as defined in the network analysis
                                                layer. The value should be specified as a
                                                comma separated list of attribute names.
                                                You can also specify a value of none to
                                                indicate that no network attributes should
                                                be used as restrictions.
        -----------------------------------     --------------------------------------------------------------------
        restrict_u_turns                        Specifies how U-Turns should be restricted in the
                                                analysis. The default is as defined in the network
                                                analysis layer. Values: esriNFSBAllowBacktrack |
                                                esriNFSBAtDeadEndsOnly | esriNFSBNoBacktrack |
                                                esriNFSBAtDeadEndsAndIntersections
        -----------------------------------     --------------------------------------------------------------------
        output_geometry_precision               The precision of the output geometry after
                                                generalization. If 0, no generalization of
                                                output geometry is performed. The default is
                                                as defined in the network service configuration.
        -----------------------------------     --------------------------------------------------------------------
        output_geometry_precision_units         The units of the output geometry precision.
                                                The default value is esriUnknownUnits.
                                                Values: esriUnknownUnits | esriCentimeters |
                                                esriDecimalDegrees | esriDecimeters |
                                                esriFeet | esriInches | esriKilometers |
                                                esriMeters | esriMiles | esriMillimeters |
                                                esriNauticalMiles | esriPoints | esriYards
        -----------------------------------     --------------------------------------------------------------------
        use_hierarchy                           If true, the hierarchy attribute for the network should be
                                                used in analysis. The default is as defined in the network
                                                layer. This cannot be used in conjunction with outputLines.
        -----------------------------------     --------------------------------------------------------------------
        time_of_day                             The date and time at the facility. If travelDirection is set
                                                to esriNATravelDirectionToFacility, the timeOfDay value
                                                specifies the arrival time at the facility. if travelDirection
                                                is set to esriNATravelDirectionFromFacility, the timeOfDay
                                                value is the departure time from the facility. The time zone
                                                for timeOfDay is specified by timeOfDayIsUTC.
        -----------------------------------     --------------------------------------------------------------------
        time_of_day_is_utc                      The time zone or zones of the timeOfDay parameter. When
                                                set to false, which is the default value, the timeOfDay
                                                parameter refers to the time zone or zones in which the
                                                facilities are located. Therefore, the start or end times
                                                of the service areas are staggered by time zone.
        -----------------------------------     --------------------------------------------------------------------
        travel_direction                        Options for traveling to or from the facility. The
                                                default is defined in the network analysis layer.
                                                Values: esriNATravelDirectionFromFacility |
                                                        esriNATravelDirectionToFacility
        -----------------------------------     --------------------------------------------------------------------
        return_z                                If true, Z values will be included in saPolygons and saPolylines
                                                geometry if the network dataset is Z-aware. The default is false.
        -----------------------------------     --------------------------------------------------------------------
        overrides                               Optional dictionary. Specify additional settings that can
                                                influence the behavior of the solver.  A list of supported
                                                override settings for each solver and their acceptable values
                                                can be obtained by contacting Esri Technical Support.
        -----------------------------------     --------------------------------------------------------------------
        preserve_objectid                       Optional Boolean.  If True, all objectid values are
                                                maintained.  The default is False.
        -----------------------------------     --------------------------------------------------------------------
        future                                  Optional boolean. If True, a future object will be returned and the process
                                                will not wait for the task to complete. The default is False, which means wait for results.
        -----------------------------------     --------------------------------------------------------------------
        ignore_invalid_locations                If true, the solver will ignore invalid
                                                locations. Otherwise, it will raise an error.
                                                Default is true.
        -----------------------------------     --------------------------------------------------------------------
        geometry_precision                      Optional Integer. Use this parameter to specify the number of decimal
                                                places in the response geometries returned by solve operation.
                                                This applies to x/y values only (not m- or z-values).
        -----------------------------------     --------------------------------------------------------------------
        geometry_precision_z                    Optional Integer. Use this parameter to specify the number of
                                                decimal places in the response geometries returned by solve operation.
                                                This applies to z values only.
        -----------------------------------     --------------------------------------------------------------------
        geometry_precision_m                    Optional Integer. Use this parameter to specify the number of
                                                decimal places in the response geometries returned by solve operation.
                                                This applies to m values only.
        ===================================     ====================================================================


        """
        if not self.properties.layerType == "esriNAServerServiceAreaLayer":
            raise TypeError(
                "The solveServiceArea operation is supported on a network "
                "layer of Service Area type only"
            )

        url = self._url + "/solveServiceArea"
        params = {"f": "json", "facilities": _handle_spatial_inputs(facilities)}

        if travel_mode:
            if isinstance(travel_mode, str):
                travel_mode = _utils.find_travel_mode(
                    gis=self._gis, travel_mode=travel_mode
                )
                params["travel_mode"] = travel_mode
            elif isinstance(travel_mode, dict):
                travel_mode = json.dumps(travel_mode)
                params["travel_mode"] = travel_mode
            else:
                travel_mode = _utils.find_travel_mode(
                    gis=self._gis, travel_mode=_utils.default_travel_mode(gis=self._gis)
                )
                params["travel_mode"] = travel_mode

        if not barriers is None:
            params["barriers"] = _handle_spatial_inputs(barriers)
        if not polyline_barriers is None:
            params["polylineBarriers"] = _handle_spatial_inputs(polyline_barriers)
        if not polygon_barriers is None:
            params["polygonBarriers"] = _handle_spatial_inputs(polygon_barriers)
        if not travel_mode is None:
            params["travelMode"] = travel_mode
        if not attribute_parameter_values is None:
            params["attributeParameterValues"] = attribute_parameter_values
        if not default_breaks is None:
            params["defaultBreaks"] = default_breaks
        if not exclude_sources_from_polygons is None:
            params["excludeSourcesFromPolygons"] = exclude_sources_from_polygons
        if not merge_similar_polygon_ranges is None:
            params["mergeSimilarPolygonRanges"] = merge_similar_polygon_ranges
        if not output_lines is None:
            params["outputLines"] = output_lines
        if not output_polygons is None:
            params["outputPolygons"] = output_polygons
        if not overlap_lines is None:
            params["overlapLines"] = overlap_lines
        if not overlap_polygons is None:
            params["overlapPolygons"] = overlap_polygons
        if not split_lines_at_breaks is None:
            params["splitLinesAtBreaks"] = split_lines_at_breaks
        if not split_polygons_at_breaks is None:
            params["splitPolygonsAtBreaks"] = split_polygons_at_breaks
        if not trim_outer_polygon is None:
            params["trimOuterPolygon"] = trim_outer_polygon
        if not trim_polygon_distance is None:
            params["trimPolygonDistance"] = trim_polygon_distance
        if not trim_polygon_distance_units is None:
            params["trimPolygonDistanceUnits"] = trim_polygon_distance_units
        if not return_facilities is None:
            params["returnFacilities"] = return_facilities
        if not return_barriers is None:
            params["returnBarriers"] = return_barriers
        if not return_polyline_barriers is None:
            params["returnPolylineBarriers"] = return_polyline_barriers
        if not return_polygon_barriers is None:
            params["returnPolygonBarriers"] = return_polygon_barriers
        if not out_sr is None:
            params["outSR"] = out_sr
        if not ignore_invalid_locations is None:
            params["ignoreInvalidLocations"] = ignore_invalid_locations
        if not accumulate_attribute_names is None:
            params["accumulateAttributeNames"] = accumulate_attribute_names
        if not impedance_attribute_name is None:
            params["impedanceAttributeName"] = impedance_attribute_name
        if not restriction_attribute_names is None:
            params["restrictionAttributeNames"] = restriction_attribute_names
        if not restrict_u_turns is None:
            params["restrictUTurns"] = restrict_u_turns
        if not output_geometry_precision is None:
            params["outputGeometryPrecision"] = output_geometry_precision
        if not output_geometry_precision_units is None:
            params["outputGeometryPrecisionUnits"] = output_geometry_precision_units
        if not use_hierarchy is None:
            params["useHierarchy"] = use_hierarchy
        if not time_of_day is None:
            if isinstance(time_of_day, datetime.datetime):
                time_of_day = f"{time_of_day.timestamp() * 1000}"
            params["timeOfDay"] = time_of_day
        if not time_of_day_is_utc is None:
            params["timeOfDayIsUTC"] = time_of_day_is_utc
        if not travel_direction is None:
            params["travelDirection"] = travel_direction
        if not return_z is None:
            params["returnZ"] = return_z
        if not overrides is None:
            params["overrides"] = overrides
        if not preserve_objectid is None:
            params["preserveObjectID"] = preserve_objectid
        if not geometry_precision is None:
            params["geometryPrecision"] = geometry_precision
        if not geometry_precision_z is None:
            params["geometryPrecisionZ"] = geometry_precision_z
        if not geometry_precision_m is None:
            params["geometryPrecisionM"] = geometry_precision_m
        if future:
            f = self._run_async(
                self._con.post,
                **{"path": url, "postdata": params},
            )
            return NAJob(future=f, task="Solve Service Area")
        return self._con.post(path=url, postdata=params)


###########################################################################
class ClosestFacilityLayer(NetworkLayer):
    """
    The Closest Facility Network Layer which has common properties of Network
    Layer as well as some attributes unique to Closest Facility Layer
    only.
    """

    def solve_closest_facility(
        self,
        incidents,
        facilities: Union[FeatureSet, Point, list, dict],
        barriers: Optional[Union[Point, FeatureSet, dict[str, Any]]] = None,
        polyline_barriers: Optional[Union[Polyline, FeatureSet, dict[str, Any]]] = None,
        polygon_barriers: Optional[Union[Polygon, FeatureSet, dict[str, Any]]] = None,
        travel_mode: Optional[str] = None,
        attribute_parameter_values: Optional[Union[str, list]] = None,
        return_directions: bool = False,
        directions_language: Optional[str] = None,
        directions_style_name: Optional[str] = None,
        directions_length_units: Optional[str] = None,
        directions_time_attribute_name: Optional[str] = None,
        return_cf_routes: bool = True,
        return_facilities: bool = False,
        return_incidents: bool = False,
        return_barriers: bool = False,
        return_polyline_barriers: bool = False,
        return_polygon_barriers: bool = False,
        output_lines: Optional[str] = None,
        default_cutoff: Optional[float] = None,
        default_target_facility_count: Optional[int] = None,
        travel_direction: Optional[str] = None,
        out_sr: Optional[int] = None,
        accumulate_attribute_names: Optional[str] = None,
        impedance_attribute_name: Optional[str] = None,
        restriction_attribute_names: Optional[str] = None,
        restrict_u_turns: Optional[str] = None,
        use_hierarchy: bool = True,
        output_geometry_precision: Optional[str] = None,
        output_geometry_precision_units: Optional[str] = None,
        time_of_day: Optional[datetime.datetime] = None,
        time_of_day_is_utc: Optional[str] = None,
        time_of_day_usage: Optional[str] = None,
        return_z: bool = False,
        overrides: Optional[dict[str, Any]] = None,
        preserve_objectid: bool = False,
        future: bool = False,
        ignore_invalid_locations: bool = True,
        directions_output_type: Optional[str] = None,
        return_traversed_edges: Optional[bool] = None,
        return_traversed_junctions: Optional[bool] = None,
        return_traversed_turns: Optional[bool] = None,
        geometry_precision: Optional[int] = None,
        geometry_precision_z: Optional[int] = None,
        geometry_precision_m: Optional[int] = None,
    ):
        """The solve operation is performed on a network layer resource of
        type closest facility (layerType is esriNAServerClosestFacilityLayer).
        You can provide arguments to the solve route operation as query
        parameters.

        ===================================     ====================================================================
        **Parameter**                            **Description**
        -----------------------------------     --------------------------------------------------------------------
        facilities                              The set of facilities loaded as network locations
                                                during analysis. Facilities can be specified using
                                                a simple comma / semi-colon based syntax or as a
                                                JSON structure. If facilities are not specified,
                                                preloaded facilities from the map document are used
                                                in the analysis. If an empty json object is passed
                                                ('{}') preloaded facilities are ignored.
        -----------------------------------     --------------------------------------------------------------------
        incidents                               The set of incidents loaded as network locations
                                                during analysis. Incidents can be specified using
                                                a simple comma / semi-colon based syntax or as a
                                                JSON structure. If incidents are not specified,
                                                preloaded incidents from the map document are used
                                                in the analysis.
        -----------------------------------     --------------------------------------------------------------------
        barriers                                The set of barriers loaded as network locations during
                                                analysis. Barriers can be specified using a simple comma
                                                / semi-colon based syntax or as a JSON structure. If
                                                barriers are not specified, preloaded barriers from the
                                                map document are used in the analysis. If an empty json
                                                object is passed ('{}') preloaded barriers are ignored.
        -----------------------------------     --------------------------------------------------------------------
        polyline_barriers                       The set of polyline barriers loaded as network
                                                locations during analysis. If polyline barriers
                                                are not specified, preloaded polyline barriers
                                                from the map document are used in the analysis.
                                                If an empty json object is passed ('{}')
                                                preloaded polyline barriers are ignored.
        -----------------------------------     --------------------------------------------------------------------
        polygonBarriers                         The set of polygon barriers loaded as network
                                                locations during analysis. If polygon barriers
                                                are not specified, preloaded polygon barriers
                                                from the map document are used in the analysis.
                                                If an empty json object is passed ('{}') preloaded
                                                polygon barriers are ignored.
        -----------------------------------     --------------------------------------------------------------------
        travel_mode                             Travel modes provide override values that help you
                                                quickly and consistently model a vehicle or mode of
                                                transportation. The chosen travel mode must be
                                                preconfigured on the network dataset that the routing
                                                service references.
        -----------------------------------     --------------------------------------------------------------------
        attribute_parameter_values              A set of attribute parameter values that
                                                can be parameterized to determine which
                                                network elements can be used by a vehicle.
        -----------------------------------     --------------------------------------------------------------------
        return_directions                       If true, directions will be generated and returned
                                                with the analysis results. Default is false.
        -----------------------------------     --------------------------------------------------------------------
        directions_language                     The language to be used when computing directions.
                                                The default is the language of the server's operating
                                                system. The list of supported languages can be found
                                                in REST layer description.
        -----------------------------------     --------------------------------------------------------------------
        directions_output_type                  Defines content, verbosity of returned
                                                directions. The default is esriDOTStandard.
                                                Values: esriDOTComplete | esriDOTCompleteNoEvents
                                                | esriDOTInstructionsOnly | esriDOTStandard |
                                                esriDOTSummaryOnly
        -----------------------------------     --------------------------------------------------------------------
        directions_style_name                   The style to be used when returning the directions.
                                                The default is as defined in the network layer. The
                                                list of supported styles can be found in REST
                                                layer description.
        -----------------------------------     --------------------------------------------------------------------
        directions_length_units                 The length units to use when computing directions.
                                                The default is as defined in the network layer.
                                                Values: esriNAUFeet | esriNAUKilometers |
                                                esriNAUMeters | esriNAUMiles |
                                                esriNAUNauticalMiles | esriNAUYards |
                                                esriNAUUnknown
        -----------------------------------     --------------------------------------------------------------------
        directions_time_attribute_name          The name of network attribute to use for
                                                the drive time when computing directions.
                                                The default is as defined in the network
                                                layer.
        -----------------------------------     --------------------------------------------------------------------
        return_cf_routes                        If true, closest facilities routes will be returned
                                                with the analysis results. Default is true.
        -----------------------------------     --------------------------------------------------------------------
        return_facilities                       If true, facilities  will be returned with the
                                                analysis results. Default is false.
        -----------------------------------     --------------------------------------------------------------------
        return_incidents                        If true, incidents will be returned with the
                                                analysis results. Default is false.
        -----------------------------------     --------------------------------------------------------------------
        return_barriers                         If true, barriers will be returned with the analysis
                                                results. Default is false.
        -----------------------------------     --------------------------------------------------------------------
        return_polyline_barriers                If true, polyline barriers will be returned
                                                with the analysis results. Default is false.
        -----------------------------------     --------------------------------------------------------------------
        return_polygon_barriers                 If true, polygon barriers will be returned with
                                                the analysis results. Default is false.
        -----------------------------------     --------------------------------------------------------------------
        output_lines                            The type of output lines to be generated in the result.
                                                The default is as defined in the network layer.
                                                Values: esriNAOutputLineTrueShape |
                                                esriNAOutputLineTrueShapeWithMeasure |
                                                esriNAOutputLineStraight | esriNAOutputLineNone
        -----------------------------------     --------------------------------------------------------------------
        default_cutoff                          The default cutoff value to stop traversing.
        -----------------------------------     --------------------------------------------------------------------
        default_target_facility_count           The default number of facilities to find.
        -----------------------------------     --------------------------------------------------------------------
        travel_direction                        Options for traveling to or from the facility.
                                                The default is defined in the network layer.
                                                Values: esriNATravelDirectionFromFacility |
                                                esriNATravelDirectionToFacility
        -----------------------------------     --------------------------------------------------------------------
        out_sr                                  The spatial reference of the geometries returned with the
                                                analysis results.
        -----------------------------------     --------------------------------------------------------------------
        accumulate_attribute_names              The list of network attribute names to be
                                                accumulated with the analysis. The default is
                                                as defined in the network layer. The value
                                                should be specified as a comma separated list
                                                of attribute names. You can also specify a
                                                value of none to indicate that no network
                                                attributes should be accumulated.
        -----------------------------------     --------------------------------------------------------------------
        impedance_attribute_name                The network attribute name to be used as the
                                                impedance attribute in analysis. The default is
                                                as defined in the network layer.
        -----------------------------------     --------------------------------------------------------------------
        restriction_attribute_names             The list of network attribute names to be
                                                used as restrictions with the analysis. The
                                                default is as defined in the network layer.
                                                The value should be specified as a comma
                                                separated list of attribute names. You can
                                                also specify a value of none to indicate that
                                                no network attributes should be used as
                                                restrictions.
        -----------------------------------     --------------------------------------------------------------------
        restrict_u_turns                        Specifies how U-Turns should be restricted in the
                                                analysis. The default is as defined in the network
                                                layer. Values: esriNFSBAllowBacktrack |
                                                esriNFSBAtDeadEndsOnly | esriNFSBNoBacktrack |
                                                esriNFSBAtDeadEndsAndIntersections
        -----------------------------------     --------------------------------------------------------------------
        use_hierarchy                           If true, the hierarchy attribute for the network should
                                                be used in analysis. The default is as defined in the
                                                network layer.
        -----------------------------------     --------------------------------------------------------------------
        output_geometry_precision               The precision of the output geometry after
                                                generalization. If 0, no generalization of
                                                output geometry is performed. The default is
                                                as defined in the network service
                                                configuration.
        -----------------------------------     --------------------------------------------------------------------
        output_geometry_precision_units         The units of the output geometry
                                                precision. The default value is
                                                esriUnknownUnits. Values: esriUnknownUnits
                                                | esriCentimeters | esriDecimalDegrees |
                                                esriDecimeters | esriFeet | esriInches |
                                                esriKilometers | esriMeters | esriMiles |
                                                esriMillimeters | esriNauticalMiles |
                                                esriPoints | esriYards
        -----------------------------------     --------------------------------------------------------------------
        time_of_day                             Arrival or departure date and time. Values: specified by
                                                number of milliseconds since midnight Jan 1st, 1970, UTC.
        -----------------------------------     --------------------------------------------------------------------
        time_of_day_is_utc                      The time zone of the timeOfDay parameter. By setting
                                                timeOfDayIsUTC to true, the timeOfDay parameter refers
                                                to Coordinated Universal Time (UTC). Choose this option
                                                if you want to find what's nearest for a specific time,
                                                such as now, but aren't certain in which time zone the
                                                facilities or incidents will be located.
        -----------------------------------     --------------------------------------------------------------------
        time_of_day_usage                       Defines the way timeOfDay value is used. The default
                                                is as defined in the network layer.
                                                Values: esriNATimeOfDayUseAsStartTime |
                                                esriNATimeOfDayUseAsEndTime
        -----------------------------------     --------------------------------------------------------------------
        return_z                                If true, Z values will be included in the returned routes and
                                                compressed geometry if the network dataset is Z-aware.
                                                The default is false.
        -----------------------------------     --------------------------------------------------------------------
        overrides                               Optional dictionary. Specify additional settings that can influence
                                                the behavior of the solver.  A list of supported override settings
                                                for each solver and their acceptable values can be obtained by
                                                contacting Esri Technical Support.
        -----------------------------------     --------------------------------------------------------------------
        preserve_objectid                       Optional Boolean.  If True, all objectid values are
                                                maintained. The default is False.
        -----------------------------------     --------------------------------------------------------------------
        future                                  Optional boolean. If True, a future object will be returned and the process
                                                will not wait for the task to complete. The default is False, which means wait for results.
        -----------------------------------     --------------------------------------------------------------------
        ignore_invalid_locations                If true, the solver will ignore invalid
                                                locations. Otherwise, it will raise an error.
                                                Default is true.
        -----------------------------------     --------------------------------------------------------------------
        return_traversed_edges                  Optional boolean. Specify whether traversed edges will be returned
                                                by the service.
        -----------------------------------     --------------------------------------------------------------------
        return_traversed_junctions              Optional boolean. Specify whether traversed junctions will be
                                                returned by the service.
        -----------------------------------     --------------------------------------------------------------------
        return_traversed_turns                  Optional boolean. Specify whether traversed turns will be returned
                                                by the service.
        -----------------------------------     --------------------------------------------------------------------
        geometry_precision                      Optional Integer. Use this parameter to specify the number
                                                of decimal places in the response geometries returned by solve operation.
                                                This applies to x/y values only (not m- or z-values).
        -----------------------------------     --------------------------------------------------------------------
        geometry_precision_z                    Optional Integer. Use this parameter specify the number of decimal
                                                places in the response geometries returned by solve operation.
                                                This applies to z-value only.
        -----------------------------------     --------------------------------------------------------------------
        geometry_precision_m                    Optional Integer. Use this parameter to specify the number of decimal
                                                places in the response geometries returned by solve operation.
                                                This applies to m-value only.
        ===================================     ====================================================================


        """

        if not self.properties.layerType == "esriNAServerClosestFacilityLayer":
            raise TypeError(
                "The solveClosestFacility operation is supported on a network "
                "layer of Closest Facility type only"
            )

        url = self._url + "/solveClosestFacility"
        params = {
            "f": "json",
            "facilities": _handle_spatial_inputs(facilities),
            "incidents": _handle_spatial_inputs(incidents),
        }

        if travel_mode:
            if isinstance(travel_mode, str):
                travel_mode = _utils.find_travel_mode(
                    gis=self._gis, travel_mode=travel_mode
                )
                params["travel_mode"] = travel_mode
            elif isinstance(travel_mode, dict):
                travel_mode = json.dumps(travel_mode)
                params["travel_mode"] = travel_mode
            else:
                travel_mode = _utils.find_travel_mode(
                    gis=self._gis, travel_mode=_utils.default_travel_mode(gis=self._gis)
                )
                params["travel_mode"] = travel_mode

        if not barriers is None:
            params["barriers"] = _handle_spatial_inputs(barriers)
        if not polyline_barriers is None:
            params["polylineBarriers"] = _handle_spatial_inputs(polyline_barriers)
        if not polygon_barriers is None:
            params["polygonBarriers"] = _handle_spatial_inputs(polygon_barriers)
        if not travel_mode is None:
            params["travelMode"] = travel_mode
        if not attribute_parameter_values is None:
            params["attributeParameterValues"] = attribute_parameter_values
        if not return_directions is None:
            params["returnDirections"] = return_directions
        if not directions_language is None:
            params["directionsLanguage"] = directions_language
        if not directions_style_name is None:
            params["directionsStyleName"] = directions_style_name
        if not directions_length_units is None:
            params["directionsLengthUnits"] = directions_length_units
        if not directions_time_attribute_name is None:
            params["directionsTimeAttributeName"] = directions_time_attribute_name
        if not directions_output_type is None:
            params["directionsOutputType"] = directions_output_type
        if not return_cf_routes is None:
            params["returnCFRoutes"] = return_cf_routes
        if not return_facilities is None:
            params["returnFacilities"] = return_facilities
        if not return_incidents is None:
            params["returnIncidents"] = return_incidents
        if not return_barriers is None:
            params["returnBarriers"] = return_barriers
        if not return_polyline_barriers is None:
            params["returnPolylineBarriers"] = return_polyline_barriers
        if not return_polygon_barriers is None:
            params["returnPolygonBarriers"] = return_polygon_barriers
        if not output_lines is None:
            params["outputLines"] = output_lines
        if not default_cutoff is None:
            params["defaultCutoff"] = default_cutoff
        if not default_target_facility_count is None:
            params["defaultTargetFacilityCount"] = default_target_facility_count
        if not travel_direction is None:
            params["travelDirection"] = travel_direction
        if not out_sr is None:
            params["outSR"] = out_sr
        if not ignore_invalid_locations is None:
            params["ignoreInvalidLocations"] = ignore_invalid_locations
        if not accumulate_attribute_names is None:
            params["accumulateAttributeNames"] = accumulate_attribute_names
        if not impedance_attribute_name is None:
            params["impedanceAttributeName"] = impedance_attribute_name
        if not restriction_attribute_names is None:
            params["restrictionAttributeNames"] = restriction_attribute_names
        if not restrict_u_turns is None:
            params["restrictUTurns"] = restrict_u_turns
        if not use_hierarchy is None:
            params["useHierarchy"] = use_hierarchy
        if not output_geometry_precision is None:
            params["outputGeometryPrecision"] = output_geometry_precision
        if not output_geometry_precision_units is None:
            params["outputGeometryPrecisionUnits"] = output_geometry_precision_units
        if not time_of_day is None:
            if isinstance(time_of_day, datetime.datetime):
                time_of_day = f"{time_of_day.timestamp() * 1000}"
            params["timeOfDay"] = time_of_day
        if not time_of_day_is_utc is None:
            params["timeOfDayIsUTC"] = time_of_day_is_utc
        if not time_of_day_usage is None:
            params["timeOfDayUsage"] = time_of_day_usage
        if not return_z is None:
            params["returnZ"] = return_z
        if not overrides is None:
            params["overrides"] = overrides
        if not preserve_objectid is None:
            params["preserveObjectID"] = preserve_objectid
        if not geometry_precision is None:
            params["geometryPrecision"] = geometry_precision
        if not geometry_precision_z is None:
            params["geometryPrecisionZ"] = geometry_precision_z
        if not geometry_precision_m is None:
            params["geometryPrecisionM"] = geometry_precision_m
        if not return_traversed_edges is None:
            params["returnTraversedEdges"] = return_traversed_edges
        if not return_traversed_junctions is None:
            params["returnTraversedJunctions"] = return_traversed_junctions
        if not return_traversed_turns is None:
            params["returnTraversedTurns"] = return_traversed_turns
        if future:
            f = self._run_async(self._con.post, **{"path": url, "postdata": params})
            return NAJob(future=f, task="Solve Closest Facility")
        return self._con.post(path=url, postdata=params)


###########################################################################
class ODCostMatrixLayer(NetworkLayer):
    """
    OD Cost Matrix Layer is part of the Network Layer services.  It allows users
    to generate cost matrix data for a given set of input.
    """

    def solve_od_cost_matrix(
        self,
        origins,
        destinations,
        default_cutoff: Optional[float] = None,
        default_target_destination_count: Optional[int] = None,
        travel_mode: Optional[str] = None,
        output_type: str = "Sparse Matrix",
        time_of_day: Optional[datetime.datetime] = None,
        time_of_day_is_utc: Optional[str] = None,
        barriers: Optional[Union[Point, FeatureSet, dict[str, Any]]] = None,
        polyline_barriers: Optional[Union[Polyline, FeatureSet, dict[str, Any]]] = None,
        polygon_barriers: Optional[Union[Polygon, FeatureSet, dict[str, Any]]] = None,
        impedance_attribute_name: Optional[str] = None,
        accumulate_attribute_names: Optional[str] = None,
        restriction_attribute_names: Optional[str] = None,
        attribute_parameter_values: Optional[str] = None,
        restrict_u_turns: Optional[str] = None,
        use_hierarchy: bool = True,
        return_origins: bool = False,
        return_destinations: bool = False,
        return_barriers: bool = False,
        return_polyline_barriers: bool = False,
        return_polygon_barriers: bool = False,
        out_sr: Optional[int] = None,
        ignore_invalid_locations: bool = True,
        return_z: bool = False,
        overrides: Optional[dict[str, Any]] = None,
        future: bool = False,
        geometry_precision: Optional[int] = None,
        geometry_precision_z: Optional[int] = None,
    ):
        """

        The Origin Destination Cost Matrix service helps you to create an
        origin-destination (OD) cost matrix from multiple origins to
        multiple destinations. An Origin Destination Cost Matrix is a table
        that contains the cost, such as the travel time or travel distance,
        from every origin to every destination. Additionally, it ranks the
        destinations that each origin connects to in ascending order based
        on the minimum cost required to travel from that origin to each
        destination. When generating an OD Cost Matrix, you can optionally
        specify the maximum number of destinations to find for each origin
        and the maximum time or distance to travel when searching for
        destinations.

        The results from the Origin Destination Cost Matrix service often
        become input for other spatial analyses where the cost to travel on
        the street network is more appropriate than straight-line cost.

        The travel time and/or distance for each origin-destination pair is
        stored in the output matrix (default) or as part of the attributes
        of the output lines, which can have no shapes or a straight line
        shape. Even though the lines are straight, they always store the
        travel time and/or travel distance based on the street network, not
        based on Euclidean distance.

        ====================================     ====================================================================
        **Parameter**                             **Description**
        ------------------------------------     --------------------------------------------------------------------
        origins                                  Required FeatureLayer/SeDF/FeatureSet.
                                                 Specifies the starting points from which to travel to the destinations.
        ------------------------------------     --------------------------------------------------------------------
        destinations                             Required FeatureLayer/SeDF/FeatureSet.
                                                 Specifies the ending point locations to travel to from the origins.
        ------------------------------------     --------------------------------------------------------------------
        default_cutoff                           Optional Float. Specify the travel time or travel distance value at
                                                 which to stop searching for destinations. The default value is
                                                 `None` which means to search until all destinations are found for
                                                 every origin. The units are the same as the impedance attribute
                                                 units.
        ------------------------------------     --------------------------------------------------------------------
        default_target_destination_count         Optional Integer. Specify the number of destinations to find per
                                                 origin. The default value is `None` which means to search until all
                                                 destinations are found for every origin.
        ------------------------------------     --------------------------------------------------------------------
        travel_mode                              Optional String. Choose the mode of transportation for the analysis.
        ------------------------------------     --------------------------------------------------------------------
        output_type                              Optional String. Specify the type of output returned by the service.
                                                 Allowed value: `Sparse Matrix` (default), `Straight Lines`, or
                                                 `No Lines`.
        ------------------------------------     --------------------------------------------------------------------
        time_of_day                              Optional Datetime. The `time_of_day` value represents the time at which
                                                 the travel begins from the input origins.
                                                 If a value of `now` is passed, the travel begins at current time.
        ------------------------------------     --------------------------------------------------------------------
        time_of_day_is_utc                       Optional Boolean. Specify the time zone or zones of the
                                                 `time_of_day` parameter. The default is as defined
                                                 in the network layer.
        ------------------------------------     --------------------------------------------------------------------
        barriers                                 Optional FeatureLayer/SeDF/FeatureSet. Specify one or more points that act as
                                                 temporary restrictions or represent additional time or distance that
                                                 may be required to travel on the underlying streets.
        ------------------------------------     --------------------------------------------------------------------
        polyline_barriers                        Optional FeatureLayer/SeDF/FeatureSet. Specify one or more lines that prohibit
                                                 travel anywhere the lines intersect the streets.
        ------------------------------------     --------------------------------------------------------------------
        polygon_barriers                         Optional FeatureLayer/SeDF/FeatureSet. Specify polygons that either prohibit
                                                 travel or proportionately scale the time or distance required to
                                                 travel on the streets intersected by the polygons.
        ------------------------------------     --------------------------------------------------------------------
        impedance_attribute_name                 Optional String. Specify the impedance. The default is as defined
                                                 in the network layer.
        ------------------------------------     --------------------------------------------------------------------
        accumulate_attribute_names               Optional String. Specify whether the service should accumulate
                                                 values other than the value specified for `impedance_attribute_names`.

                                                 The default is as defined in the network layer. The parameter value
                                                 should be specified as a comma-separated list of names.
        ------------------------------------     --------------------------------------------------------------------
        restriction_attribute_names              Optional String. Specify which restrictions should be honored by the service.
        ------------------------------------     --------------------------------------------------------------------
        attribute_parameter_values               Optional String. Specify additional values required by an attribute or restriction.
        ------------------------------------     --------------------------------------------------------------------
        restrict_u_turns                         Optional String. Restrict or permit the route from making U-turns at
                                                 junctions. The default is as defined in the network layer.

                                                 Values: esriNFSBAllowBacktrack | esriNFSBAtDeadEndsOnly |
                                                         esriNFSBNoBacktrack | esriNFSBAtDeadEndsAndIntersections
        ------------------------------------     --------------------------------------------------------------------
        use_hierarchy                            Optional Boolean. Specify whether hierarchy should be used when finding the shortest paths. The default value is true.
        ------------------------------------     --------------------------------------------------------------------
        return_origins                           Optional Boolean. Specify whether origins will be returned by the service. The default value is false.
        ------------------------------------     --------------------------------------------------------------------
        return_destinations                      Optional Boolean. Specify whether origins will be returned by the service. The default value is false.
        ------------------------------------     --------------------------------------------------------------------
        return_barriers                          Optional Boolean. Specify whether barriers will be returned by the service. The default value is false.
        ------------------------------------     --------------------------------------------------------------------
        return_polyline_barriers                 Optional Boolean. Specify whether polyline barriers will be returned by the service. The default value is false.
        ------------------------------------     --------------------------------------------------------------------
        return_polygon_barriers                  Optional Boolean. Specify whether polygon barriers will be returned by the service. The default value is false.
        ------------------------------------     --------------------------------------------------------------------
        out_sr                                   Optional Integer. Specify the spatial reference of the geometries.
        ------------------------------------     --------------------------------------------------------------------
        ignore_invalid_locations                 Optional Boolean. Specify whether invalid input locations should be
                                                 ignored when finding the best solution. The default is True.
        ------------------------------------     --------------------------------------------------------------------
        return_z                                 Optional Boolean. Include z values for the returned geometries if supported by the underlying network. The default value is false.
        ------------------------------------     --------------------------------------------------------------------
        overrides                                Optional Dict. Specify additional settings that can influence the behavior of the solver.
        ------------------------------------     --------------------------------------------------------------------
        future                                   Optional boolean. If True, a future object will be returned and the process
                                                 will not wait for the task to complete. The default is False, which means wait for results.
        ------------------------------------     --------------------------------------------------------------------
        geometry_precision                       Optional Integer. Use this parameter to specify the number of decimal places in the response geometries returned by solve operation. This applies to x/y values only (not m- or z-values).
        ------------------------------------     --------------------------------------------------------------------
        geometry_precision_z                     Optional Integer. Use this parameter specify the number of decimal places in the response geometries returned by solve operation. This applies to z-value only.
        ====================================     ====================================================================

        :return: Dictionary or `NAJob` when `future=True`

        """
        if not self.properties.layerType == "esriNAServerODCostMatrixLayer":
            raise TypeError(
                "The solveODCostMatrix operation is supported on a network "
                "layer of OD Cost Matrix type only"
            )
        url = f"{self._url}/solveODCostMatrix"
        params = {
            "f": "json",
            "origins": _handle_spatial_inputs(origins),
            "destinations": _handle_spatial_inputs(destinations),
        }
        if travel_mode:
            if isinstance(travel_mode, str):
                travel_mode = _utils.find_travel_mode(
                    gis=self._gis, travel_mode=travel_mode
                )
                params["travel_mode"] = travel_mode
            elif isinstance(travel_mode, dict):
                travel_mode = json.dumps(travel_mode)
                params["travel_mode"] = travel_mode
            else:
                travel_mode = _utils.find_travel_mode(
                    gis=self._gis, travel_mode=_utils.default_travel_mode(gis=self._gis)
                )
                params["travel_mode"] = travel_mode
        allowed_output_types = {
            "esriNAODOutputSparseMatrix": "esriNAODOutputSparseMatrix",
            "Sparse Matrix": "esriNAODOutputSparseMatrix",
            "esriNAODOutputStraightLines": "esriNAODOutputStraightLines",
            "Straight Lines": "esriNAODOutputStraightLines",
            "esriNAODOutputNoLines": "esriNAODOutputNoLines",
            "No Lines": "esriNAODOutputNoLines",
        }
        if not default_cutoff is None:
            params["defaultCutoff"] = default_cutoff
        if not default_target_destination_count is None:
            params["defaultTargetDestinationCount"] = default_target_destination_count
        if not travel_mode is None:
            params["travelMode"] = travel_mode
        if not output_type is None:
            assert output_type in allowed_output_types
            params["outputType"] = allowed_output_types[output_type]
        if not time_of_day is None:
            if isinstance(time_of_day, datetime.datetime):
                time_of_day = f"{time_of_day.timestamp() * 1000}"
            params["timeOfDay"] = time_of_day
        if not time_of_day_is_utc is None:
            params["timeOfDayIsUTC"] = time_of_day_is_utc
        if not barriers is None:
            params["barriers"] = _handle_spatial_inputs(barriers)
        if not polyline_barriers is None:
            params["polylineBarriers"] = _handle_spatial_inputs(polyline_barriers)
        if not polygon_barriers is None:
            params["polygonBarriers"] = _handle_spatial_inputs(polygon_barriers)
        if not impedance_attribute_name is None:
            params["impedanceAttributeName"] = impedance_attribute_name
        if not accumulate_attribute_names is None:
            params["accumulateAttributeNames"] = accumulate_attribute_names
        if not restriction_attribute_names is None:
            params["restrictionAttributeNames"] = restriction_attribute_names
        if not attribute_parameter_values is None:
            params["attributeParameterValues"] = attribute_parameter_values
        if not restrict_u_turns is None:
            params["restrictUTurns"] = restrict_u_turns
        if not use_hierarchy is None:
            params["useHierarchy"] = use_hierarchy
        if not return_origins is None:
            params["returnOrigins"] = return_origins
        if not return_destinations is None:
            params["returnDestinations"] = return_destinations
        if not return_barriers is None:
            params["returnBarriers"] = return_barriers
        if not return_polyline_barriers is None:
            params["returnPolylineBarriers"] = return_polyline_barriers
        if not return_polygon_barriers is None:
            params["returnPolygonBarriers"] = return_polygon_barriers
        if not out_sr is None:
            params["outSR"] = out_sr
        if not ignore_invalid_locations is None:
            params["ignoreInvalidLocations"] = ignore_invalid_locations
        if not return_z is None:
            params["returnZ"] = return_z
        if not overrides is None:
            params["overrides"] = overrides
        if not geometry_precision is None:
            params["geometryPrecision"] = geometry_precision
        if not geometry_precision_z is None:
            params["geometryPrecisionZ"] = geometry_precision_z

        if future:
            f = self._run_async(self._con.post, **{"path": url, "postdata": params})
            return NAJob(future=f, task="Solve OD Cost Matrix")
        return self._con.post(path=url, postdata=params)

    # ----------------------------------------------------------------------
    def retrieve_travel_modes(self):
        """
        Identify all the valid travel modes that have been defined on the
        network dataset or in the portal if the GIS server is federated

        :return: Dictionary

        """
        from arcgis._impl.common._isd import InsensitiveDict

        url = self._url + "/retrieveTravelModes"
        params = {"f": "json"}
        return InsensitiveDict(self._con.get(path=url, params=params))


###########################################################################
class NetworkDataset(_GISResource):
    """
    A network dataset containing a collection of network layers including route layers,
    service area layers and closest facility layers.
    """

    def __init__(self, url, gis=None):
        if gis is None:
            from arcgis.env import active_gis

            gis = active_gis
        if gis:
            url = _validate_url(url, gis)
        super(NetworkDataset, self).__init__(url, gis)

        try:
            from ..gis.server._service._adminfactory import AdminServiceGen

            self.service = AdminServiceGen(service=self, gis=gis)
        except:
            pass
        self._closestFacilityLayers = []
        self._routeLayers = []
        self._serviceAreaLayers = []
        self._odCostMatrix = []
        self._load_layers()

    @classmethod
    def fromitem(cls, item):
        """Creates a network dataset from a 'Network Analysis Service' Item in the GIS"""
        if not item.type == "Network Analysis Service":
            raise TypeError(
                "item must be a type of Network Analysis Service, not " + item.type
            )
        url = _validate_url(item.url, item._gis)
        return cls(url, item._gis)

    # ----------------------------------------------------------------------
    def _load_layers(self):
        """loads the various layer types"""
        self._closestFacilityLayers = []
        self._routeLayers = []
        self._serviceAreaLayers = []
        self._odCostMatrix = []
        params = {
            "f": "json",
        }
        json_dict = self._con.get(path=self._url, params=params)
        for k, v in json_dict.items():
            if k == "routeLayers" and json_dict[k]:
                self._routeLayers = []
                for rl in v:
                    self._routeLayers.append(
                        RouteLayer(url=self._url + "/%s" % rl, gis=self._gis)
                    )
            elif k == "serviceAreaLayers" and json_dict[k]:
                self._serviceAreaLayers = []
                for sal in v:
                    self._serviceAreaLayers.append(
                        ServiceAreaLayer(url=self._url + "/%s" % sal, gis=self._gis)
                    )
            elif k == "closestFacilityLayers" and json_dict[k]:
                self._closestFacilityLayers = []
                for cf in v:
                    self._closestFacilityLayers.append(
                        ClosestFacilityLayer(url=self._url + "/%s" % cf, gis=self._gis)
                    )
            elif k == "odCostMatrixLayers" and json_dict[k]:
                self._odCostMatrix = []
                for cf in v:
                    self._odCostMatrix.append(
                        ODCostMatrixLayer(url=self._url + "/%s" % cf, gis=self._gis)
                    )

    # ----------------------------------------------------------------------
    @property
    def route_layers(self):
        """List of route layers in this network dataset"""
        if self._routeLayers is None:
            self._load_layers()
        return self._routeLayers

    # ----------------------------------------------------------------------
    @property
    def service_area_layers(self):
        """List of service area layers in this network dataset"""
        if self._serviceAreaLayers is None:
            self._load_layers()
        return self._serviceAreaLayers

    # ----------------------------------------------------------------------
    @property
    def closest_facility_layers(self):
        """List of closest facility layers in this network dataset"""
        if self._closestFacilityLayers is None:
            self._load_layers()
        return self._closestFacilityLayers

    @property
    def od_cost_matrix_layers(self):
        """
        List of OD Cost Matrix Layers

        :return: List
        """
        if self._odCostMatrix is None:
            self._load_layers()
        return self._odCostMatrix



from __future__ import annotations

import requests
from enum import Enum
from typing import Any, Iterator
from functools import lru_cache
from arcgis.gis import GIS
from arcgis.auth import EsriSession

__all__ = ["get_places_api"]


###########################################################################
class PlaceIdEnums(Enum):
    """
    When obtaining a single place by ID, the enumerations can filter and
    return specific information about a specific place ID.  To obtain a
    place ID, you must perform a query first.
    """

    ALL = "all"
    ADDITIONALLOCATIONS = "additionalLocations"
    ADDITIONALLOCATIONS_DROPOFF = "additionalLocations:dropOff"
    ADDITIONALLOCATIONS_FRONTDOOR = "additionalLocations:frontDoor"
    ADDITIONALLOCATIONS_ROAD = "additionalLocations:road"
    ADDITIONALLOCATIONS_ROOF = "additionalLocations:roof"
    ADDRESS = "address"
    ADDRESS_ADMINREGION = "address:adminRegion"
    ADDRESS_CENSUSBLOCKID = "address:censusBlockId"
    ADDRESS_COUNTRY = "address:country"
    ADDRESS_DESIGNATEDMARKETAREA = "address:designatedMarketArea"
    ADDRESS_EXTENDED = "address:extended"
    ADDRESS_LOCALITY = "address:locality"
    ADDRESS_NEIGHBORHOOD = "address:neighborhood"
    ADDRESS_POBOX = "address:poBox"
    ADDRESS_POSTCODE = "address:postcode"
    ADDRESS_POSTTOWN = "address:postTown"
    ADDRESS_REGION = "address:region"
    ADDRESS_STREETADDRESS = "address:streetAddress"
    CATEGORIES = "categories"
    CONTACTINFO = "contactInfo"
    CONTACTINFO_EMAIL = "contactInfo:email"
    CONTACTINFO_FAX = "contactInfo:fax"
    CONTACTINFO_TELEPHONE = "contactInfo:telephone"
    CONTACTINFO_WEBSITE = "contactInfo:website"
    CHAINS = "chains"
    DESCRIPTION = "description"
    HOURS = "hours"
    HOURS_OPENING = "hours:opening"
    HOURS_OPENINGTEXT = "hours:openingText"
    HOURS_POPULAR = "hours:popular"
    LOCATION = "location"
    NAME = "name"
    RATING = "rating"
    RATING_PRICE = "rating:price"
    RATING_USER = "rating:user"
    SOCIALMEDIA = "socialMedia"
    SOCIALMEDIA_FACEBOOKID = "socialMedia:facebookId"
    SOCIALMEDIA_INSTAGRAM = "socialMedia:instagram"
    SOCIALMEDIA_TWITTER = "socialMedia:twitter"


###########################################################################
class PlacesAPI:
    """
    The places service is a ready-to-use location service that can search
    for businesses and geographic locations around the world. It allows
    you to find, locate, and discover detailed information about each place.
    """

    _gis: GIS = None
    _urls: dict = None

    # ---------------------------------------------------------------------
    def __init__(self, gis: GIS) -> None:
        assert gis._portal.is_arcgisonline, "The GIS must be ArcGIS Online."
        assert not gis.users.me is None or gis.properties.get(
            "appInfo", None
        ), "You must be signed into the GIS to use the Places API"
        assert self._check_privileges(
            gis=gis
        ), "The current GIS does not have PlaceAPI permissions"
        self._gis = gis
        self._urls = {
            "base_url": "https://places-api.arcgis.com/arcgis/rest/services/places-service/v1",
            "near-point": "/places/near-point",
            "within-extent": "/places/within-extent",
            "categories": "/categories",
            "places": "/places",
        }
        self.session: EsriSession = gis._con._session

    @lru_cache(maxsize=255)
    def _check_privileges(self, gis: GIS) -> bool:
        """
        Checks to see if the current login has the proper permisions to use
        the Places API

        :return: bool

        """
        user: "User" = gis.users.me
        properties: dict[str, Any] = dict(gis.properties)
        if user:
            if "premium:user:places" in user.privileges:
                return True
        else:
            if "premium:user:places" in properties.get("appInfo", {}).get(
                "privileges", []
            ):
                return True
        return False

    # ---------------------------------------------------------------------
    def __repr__(self) -> str:
        return f"< {self.__class__.__name__} @ {self._urls['base_url']} >"

    # ---------------------------------------------------------------------
    def __str__(self) -> str:
        return f"< {self.__class__.__name__} @ {self._urls['base_url']} >"

    # ---------------------------------------------------------------------
    def examine_category(self, category: str) -> dict[str, Any]:
        """
        Get the category details for a category ID.

        ======================     ===============================================================
        **Parameter**               **Description**
        ----------------------     ---------------------------------------------------------------
        category                   Required String. The category ID to examine.
        ======================     ===============================================================

        :return: Dictionary
        """
        url: str = f"{self._urls['base_url']}{self._urls['categories']}/{category}"
        params = {
            "f": "json",
        }
        resp: requests.Response = self.session.get(url=url, params=params)
        resp.raise_for_status()
        data: dict[str, Any] = resp.json()
        return data

    # ---------------------------------------------------------------------
    def find_category(self, query: str) -> dict[str, Any]:
        """
        Return the name and category ID of all categories, or categories
        which satisfy a filter.

        A category describes a type of place, such as "movie theater" or
        "zoo". The places service has over 1,000 categories (or types) of
        place. The categories fall into ten general groups: Arts and
        Entertainment, Business and Professional Services, Community and
        Government, Dining and Drinking, Events, Health and Medicine,
        Landmarks and Outdoors, Retail, Sports and Recreation, and Travel
        and Transportation.


        ======================     ===============================================================
        **Parameter**               **Description**
        ----------------------     ---------------------------------------------------------------
        query                      Required String. The filter string used to find the matching
                                   categories.
        ======================     ===============================================================

        :return: Dictionary

        """
        url: str = f"{self._urls['base_url']}{self._urls['categories']}"
        params = {
            "f": "json",
            "filter": query,
            "token": self.session.auth.token,
        }
        resp: requests.Response = self.session.get(url=url, params=params)
        resp.raise_for_status()
        data: dict[str, Any] = resp.json()
        return data

    # ---------------------------------------------------------------------
    def search_by_radius(
        self,
        point: list[float] | list[int],
        radius: float | int = 1000,
        categories: list[str] | None = None,
        search_text: str | None = None,
        page_size: int = 10,
    ) -> Iterator[dict[str, Any]]:
        """
        Search for places near a point or location by radius.

        ======================     ===============================================================
        **Parameter**               **Description**
        ----------------------     ---------------------------------------------------------------
        point                      list/tuple[float]. The X/Y coordinates centroid to search by
                                   radius by.  The coordinates must be in WGS-1984 (Lat/Long).

                                   Example: [-73.991997,40.743648]
        ----------------------     ---------------------------------------------------------------
        categories                 Optional list[str]. The category IDs to examine.
        ----------------------     ---------------------------------------------------------------
        search_text                Optional str. The free search text for places against names, categories etc.
        ----------------------     ---------------------------------------------------------------
        page_size                  Optional Integer. The amount of records to return per query. The default is 10.
        ======================     ===============================================================

        :yield: dict[str,Any]

        """
        x, y = point
        params: dict[str, Any] = {
            "f": "json",
            "x": x,
            "y": y,
            "radius": radius,
            "searchText": search_text or "",
            "pageSize": page_size,
        }
        if search_text is None:
            search_text = ""
        if not categories is None:
            params["categoriesIds"] = ",".join(categories)
        url: str = f"{self._urls['base_url']}{self._urls['near-point']}"
        resp: requests.Response = self.session.get(url=url, params=params)
        resp.raise_for_status()
        data: dict[str, Any] = resp.json()
        for result in data.get("results", []):
            yield result
        while (
            not data.get("pagination", {}).get("nextUrl") is None
            and len(data.get("results", [])) != 0
        ):
            url = data.get("pagination", {}).get("nextUrl")
            resp = self.session.get(url=url)
            resp.raise_for_status()
            data: dict[str, Any] = resp.json()
            if len(data.get("results", [])) == 0:
                return
            for result in data.get("results", []):
                yield result

    # ---------------------------------------------------------------------
    def search_by_extent(
        self,
        bbox: list[float] | list[int],
        categories: list[str] | None = None,
        search_text: str | None = None,
        page_size: int = 10,
    ) -> Iterator[dict[str, Any]]:
        """
        Search for places within an extent (bounding box).

        ======================     ===============================================================
        **Parameter**               **Description**
        ----------------------     ---------------------------------------------------------------
        bbox                       list/tuple[float]. The min X/Y and max X/Y coordinates to
                                   search within. Coordinates must be in WGS-1984 (Lat/Long).

                                   Example: [-54,-75,54,75]
        ----------------------     ---------------------------------------------------------------
        categories                 Optional list[str]. The category IDs to examine.
        ----------------------     ---------------------------------------------------------------
        search_text                Optional str. The free search text for places against names, categories etc.
        ----------------------     ---------------------------------------------------------------
        page_size                  Optional Integer. The amount of records to return per query. The default is 10.
        ======================     ===============================================================

        :return: Iterator[dict[str,Any]]

        """
        data: dict[str, Any]
        resp: requests.Response
        xmin: float | int
        ymin: float | int
        xmax: float | int
        ymax: float | int

        if search_text is None:
            search_text = ""
        xmin, ymin, xmax, ymax = bbox
        params: dict[str, Any] = {
            "f": "json",
            "xmin": xmin,
            "ymin": ymin,
            "xmax": xmax,
            "ymax": ymax,
            "searchText": search_text,
            "pageSize": page_size,
        }
        if not categories is None:
            params["categoriesIds"] = ",".join(categories)
        url: str = f"{self._urls['base_url']}{self._urls['within-extent']}"
        resp: requests.Response = self.session.get(url=url, params=params)
        resp.raise_for_status()
        data: dict[str, Any] = resp.json()
        for result in data.get("results", []):
            yield result
        while (
            not data.get("pagination", {}).get("nextUrl") is None
            and len(data.get("results", [])) != 0
        ):
            url = data.get("pagination", {}).get("nextUrl")
            resp = self.session.get(url=url)
            resp.raise_for_status()
            data: dict[str, Any] = resp.json()
            if len(data.get("results", [])) == 0:
                return
            for result in data.get("results", []):
                yield result

    # ---------------------------------------------------------------------
    def get_place_by_id(
        self,
        placeid: str,
        filters: list[PlaceIdEnums] | None = None,
    ) -> dict[str, Any]:
        """
        Get place details including name, address, description, and other attributes.

        ======================     ===============================================================
        **Parameter**               **Description**
        ----------------------     ---------------------------------------------------------------
        placeid                    Required String. The Id of the place for which you want to fetch additional details.
        ----------------------     ---------------------------------------------------------------
        fields                     Optional list[string]. The array of fields that define the attributes to return for a place.
        ======================     ===============================================================

        :returns: dict[str,Any]

        """
        if filters is None:
            filters = [PlaceIdEnums.ALL]
        url: str = f"{self._urls['base_url']}{self._urls['places']}/{placeid}"
        filters: list[str] = [f.value for f in filters if isinstance(f, PlaceIdEnums)]
        params = {
            "f": "json",
            "requestedFields": ",".join(filters),
        }
        resp: requests.Response = self.session.get(url=url, params=params)
        resp.raise_for_status()
        return resp.json()


###########################################################################


# -------------------------------------------------------------------------
@lru_cache(maxsize=50)
def get_places_api(gis: GIS) -> PlacesAPI:
    """
    Returns the PlacesAPI class for a given GIS object

    :return:
        An instance of the :class:`~arcgis.geocoding.PlacesAPI` for the
        :class:`~arcgis.gis.GIS`
    """
    return PlacesAPI(gis=gis)



"""
Business analyst provides a single interface for interacting
with ArcGIS Business Analyst - whether accessing Business Analyst
functionality locally or remotely. Local access is through
ArcGIS Pro with the Business Analyst extension and locally installed
data. Remote access is through a connection to a Web GIS through a GIS
object instance providing access to either ArcGIS Enterprise with
Business Analyst or ArcGIS Online.

.. note::

    Accessing enrich using ArcGIS Online *does* consume credits.

"""
from ._main import BusinessAnalyst, Country

__all__ = ["BusinessAnalyst", "Country"]



DEFAULT_WEBSCENE_TEXT_PROPERTY = {
    "operationalLayers": [],
    "baseMap": {
        "id": "basemap",
        "title": "Topographic Vector",
        "baseMapLayers": [
            {
                "id": "world-hillshade-layer",
                "url": "//services.arcgisonline.com/arcgis/rest/services/Elevation/World_Hillshade/MapServer",
                "layerType": "ArcGISTiledMapServiceLayer",
                "title": "World Hillshade",
                "showLegend": False,
                "visibility": True,
                "opacity": 1,
            },
            {
                "id": "topo-vector-base-layer",
                "styleUrl": "//cdn.arcgis.com/sharing/rest/content/items/7dc6cea0b1764a1f9af2e679f642f0f5/resources/styles/root.json",
                "layerType": "VectorTileLayer",
                "title": "World Topo",
                "visibility": True,
                "opacity": 1,
            },
        ],
        "elevationLayers": [
            {
                "url": "https://elevation3d.arcgis.com/arcgis/rest/services/WorldElevation3D/Terrain3D/ImageServer",
                "id": "globalElevation",
                "listMode": "hide",
                "layerType": "ArcGISTiledElevationServiceLayer",
                "title": "Elevation",
            }
        ],
    },
    "ground": {
        "layers": [
            {
                "url": "https://elevation3d.arcgis.com/arcgis/rest/services/WorldElevation3D/Terrain3D/ImageServer",
                "id": "globalElevation",
                "listMode": "hide",
                "layerType": "ArcGISTiledElevationServiceLayer",
                "title": "Elevation",
            }
        ]
    },
    "viewingMode": "global",
    "spatialReference": {"latestWkid": 3857, "wkid": 102100},
    "version": "1.11",
    "authoringApp": "PortalMycontentCreate",
    "authoringAppVersion": "6.2.0.0",
}



from __future__ import annotations
from typing import Any
from arcgis import env
from arcgis._impl.common._mixins import PropertyMap
from arcgis.features._utility import TraceConfigurationsManager
from arcgis.features._trace_configuration import TraceConfiguration
from arcgis.auth.tools import LazyLoader

geometry = LazyLoader("arcgis.geometry")


########################################################################
class TraceNetworkManager(object):
    """
    The Trace Network Service exposes analytic capabilities (tracing)
    as well as validation of network topology.

    =====================   ===========================================
    **Inputs**              **Description**
    ---------------------   -------------------------------------------
    url                     Required String. The web endpoint to the trace service.
    ---------------------   -------------------------------------------
    version                 Optional Version. The `Version` class where the branch version will take place.
    ---------------------   -------------------------------------------
    gis                     Optional GIS. The `GIS` connection object.
    =====================   ===========================================


    """

    _con = None
    _gis = None
    _url = None
    _version = None
    _property = None
    _version_guid = None
    _version_name = None

    # ----------------------------------------------------------------------
    def __init__(self, url, version=None, gis=None):
        """Constructor"""
        if gis is None:
            gis = env.active_gis
        self._gis = gis
        self._con = gis._portal.con
        self._url = url
        if version:
            self._version = version
            self._version_guid = version._guid
            self._version_name = version.properties.versionName
        else:
            self._version = None
            self._version_guid = None
            self._version_name = None

    # ----------------------------------------------------------------------
    def _init(self):
        """initializer"""
        try:
            res = self._con.get(self._url, {"f": "json"})
            self._property = PropertyMap(res)
        except Exception as e:
            self._property = PropertyMap({})

    # ----------------------------------------------------------------------
    @property
    def properties(self):
        """returns the properties for the service"""
        if self._property is None:
            self._init()
        return self._property

    # ----------------------------------------------------------------------
    def trace(
        self,
        locations: list[dict],
        trace_type: str,
        moment: str | None = None,
        configuration: dict | TraceConfiguration | None = None,
        result_types: list[dict] | None = None,
        run_async: bool = False,
    ) -> dict:
        """
        A trace refers to a preconfigured algorithm that systematically
        travels a network to return results. Multiple parameters and properties
        are provided with the trace operation that support various analytic workflows.
        All traces use the network topology to read cached information about network features.
        This can improve performance of complex traces on large networks.
        Trace results are not guaranteed to accurately represent a trace network when
        dirty areas are present. The network topology must be validated to ensure that it
        reflects the most recent edits or updates made to the network.

        .. note::
            The active portal account must be licensed with the ArcGIS Trace
            Network user type extention to use this operation.

        ====================    ==================================================
        **Parameter**           **Description**
        --------------------    --------------------------------------------------
        locations               Required list of dictionaries. The locations for
                                starting points and barriers. An empty array must
                                be used when performing a subnetwork trace if a
                                subnetworkName is provided as part of the
                                `configuration` - for example, `locations=[]`.


                                The location is ignored by the trace if the following
                                required properties are not defined:
                                * `percentAlong` : required for edge features and objects.


                                .. code-block:: python

                                    [{
                                        "traceLocationType" : "startingPoint" | "barrier",
                                        "globalId" : <guid>,
                                        “percentAlong” : <double>, // optional
                                    }]
        --------------------    --------------------------------------------------
        trace_type              Required string. Specifies the core algorithm that
                                will be executed to analyze the network. Can be
                                configured using the `configuration` parameter.

                                `Values: 'connected' | 'subnetwork' | 'subnetworkController' |
                                'upstream' | 'downstream' | 'loops' | 'shortestPath' |
                                'isolation'`
        --------------------    --------------------------------------------------
        moment                  Optional Integer. Specifies the session moment. This
                                should only be specified if you do not want to use
                                the current moment.

                                Example: moment = <Epoch time in milliseconds>
        --------------------    --------------------------------------------------
        configuration           Optional dictionary or instance of TraceConfiguration
                                class. Specifies the collection of
                                trace configuration properties. Depending on the
                                `trace_type`, some properties are required.

                                To see all configuration properties see:
                                `Trace Configuration Properties
                                <https://developers.arcgis.com/rest/services-reference/enterprise/trace-trace-network-server-.htm#GUID-F0C932FD-B403-4223-9B00-E44D156C7DF9/>`_
        --------------------    --------------------------------------------------
        result_types            Optional parameter specifying hte types of results
                                to return.

                                .. code-block::
                                    [{
                                        "type" : "elements" | "aggregatedGeometry" | "connectivity",
                                        "includeGeometry" : true | false,
                                        "includePropagatedValues": true | false,
                                        "networkAttributeNames" :["attribute1Name","attribute2Name",...],
                                        "diagramTemplateName": <value>,
                                        "resultTypeFields":[{"networkSourceId":<long>,"fieldname":<value>},...]
                                    },...]

                                .. note::
                                    ArcGIS Enterprise 10.9.1 or later is required when
                                    using the `connectivity` type.
        ====================    ==================================================

        :return: If `asynchronous = True` then the status URL is returned for the job. Otherwise,
        a Dictionary of the Trace Results is returned.
        """
        url = "%s/trace" % self._url
        if isinstance(configuration, TraceConfiguration):
            configuration = configuration.to_dict()
        params = {
            "f": "json",
            "traceType": trace_type,
            "moment": moment,
            "traceLocations": locations,
            "traceConfiguration": configuration,
            "resultTypes": result_types,
            "async": run_async,
        }
        return self._con.post(url, params)

    # ----------------------------------------------------------------------
    def query_network_moments(
        self,
        moments_to_return: list[str] = ["all"],
        moment: str | None = None,
    ) -> dict:
        """
        The `query_network_moments` operation returns the moments related
        to the network topology and operations against the topology. This
        includes when the topology was initially enabled, when it was last
        validated, when the topology was last disabled (and later enabled),
        and when the definition of the trace network was last modified.

        ====================================        ====================================================================
        **Parameter**                                **Description**
        ------------------------------------        --------------------------------------------------------------------
        moments_to_return                           Optional List of Strings. Represents the collection of validate moments to
                                                    return. Default is all.

                                                    `Values: ["initialEnableTopology" | "fullValidateTopology" |
                                                            "partialValidateTopology" | "enableTopology" | "disableTopology" |
                                                            "definitionModification" | "indexUpdate" | "all" ]`
        ------------------------------------        --------------------------------------------------------------------
        moment                                      Optional Epoch in Time in Seconds.
                                                    Example: `moment=1603109606`
        ====================================        ====================================================================

        """
        url = "%s/queryNetworkMoments" % self._url
        params = {
            "f": "json",
            "gdbVersion": self._version_name,
            "sessionId": self._version_guid,
            "momentsToReturn": moments_to_return,
            "moment": moment,
        }
        return self._con.post(url, params)

    # ----------------------------------------------------------------------
    def validate_topology(
        self,
        envelope: dict[str, Any] | geometry.Envelope,
        return_edits: bool = False,
    ) -> dict:
        """
        Validating the network topology for a trace network maintains
        consistency between feature editing space and network topology space.
        Validating a network topology may include all or a subset of the
        dirty areas present in the network. Validation of network topology
        is supported synchronously and asynchronously.

        ====================================        ====================================================================
        **Parameter**                                **Description**
        ------------------------------------        --------------------------------------------------------------------
        envelope                                    Required Dictionary or Envelope. The envelope of the area to validate.

                                                    .. code-block:: python

                                                        {
                                                            "xmin": <minimum x-coordinate>,
                                                            "ymin": <minimum y-coordinate>,
                                                            "xmax": <maximum x-coordinate>,
                                                            "ymax": <maximum y-coordinate>,
                                                            "spatialReference": {
                                                            "wkid": <spatial reference well-known identifier>,
                                                            "latestWkid": <the current wkid value associated with the wkid>
                                                            }
                                                        }
        ------------------------------------        --------------------------------------------------------------------
        return_edits                                Optional Boolean. Returned results are organized in a layer-by-layer fashion.
                                                    If `return_edits` is set to True, each layer may have edited features
                                                    returned in an editedFeatures object.
                                                    The editedFeatures object returns full features including the original
                                                    features prior to delete; the original and current features for updates;
                                                    and the current rows for inserts, which may contain implicit changes
                                                    (for example, as a result of a calculation rule).

                                                    The response includes no editedFeatures and 'exceededTransferLimit = true'
                                                    if the count of edited features to return is more than the maxRecordCount.
                                                    If clients are using this parameter to maintain a cache, they should
                                                    invalidate the cache when exceededTransferLimit = true is returned.
                                                    If the server encounters an error when generating the list
                                                    of edits is the response, exceededTransferLimit = true is also returned.

                                                    Edited features are returned in the spatial reference
                                                    of the feature service as defined by the service's spatialReferenceobject
                                                    or by the spatialReference of the layer's extent object.
        ====================================        ====================================================================

        :return: Dictionary indicating 'success' or 'error'

        """
        url = "%s/validateNetworkTopology" % self._url
        params = {
            "f": "json",
            "gdbVersion": self._version_name,
            "sessionId": self._version_guid,
            "validateArea": envelope,
            "returnEdits": return_edits,
        }
        return self._con.post(url, params)

    # ----------------------------------------------------------------------
    def trace_configurations(self) -> TraceConfigurationsManager:
        """
        The `trace_configurations` resource provides access to all trace
        configuration operations for a trace network.
        It is returned as an array of named trace configurations with the creator,
        name, and global ID for each.
        """

        if self._gis.version >= [9, 2]:
            url = "%s/traceConfigurations" % self._url
            return TraceConfigurationsManager(
                url, version=self._version, gis=self._gis, service_url=self._url
            )
